import * as React from "react";
import { GraphNodeState, ICanvasAnnotation, TSetData } from "../components";
import { IRectConfig, ITheme } from "../contexts";
import { useTheme } from "../hooks";
import { getRectHeight, getRectWidth, updateAnnotation } from "../utils";

interface IProps {
  annotation: ICanvasAnnotation;
  config: IRectConfig<ICanvasAnnotation>;
  setData: TSetData;
}

const getChangeHandler = (
  annotation: ICanvasAnnotation,
  setAnnotations: TSetData
) => (e: React.ChangeEvent<HTMLTextAreaElement>) => {
  updateAnnotation(
    annotation.id,
    { data: { ...annotation.data, comment: e.target.value } },
    setAnnotations
  );
};

const getContainerStyles = (
  annotation: ICanvasAnnotation,
  theme: ITheme,
  config: IRectConfig<ICanvasAnnotation>
): React.CSSProperties => {
  let borderColor = theme.annotationBgColor;
  switch (annotation.state) {
    case GraphNodeState.activated:
      borderColor = theme.borderColor;
      break;
    case GraphNodeState.selected:
      borderColor = theme.primaryColor;
      break;
    default:
  }

  return {
    position: "relative",
    boxSizing: "border-box",
    height: getRectHeight<ICanvasAnnotation>(config, annotation),
    width: getRectWidth<ICanvasAnnotation>(config, annotation),
    padding: 8,
    backgroundColor: theme.annotationBgColor,
    borderRadius: 2,
    cursor: "move",
    border: `1px solid ${borderColor}`
  };
};

export const AnnotationContent: React.FunctionComponent<IProps> = props => {
  const { annotation, config, setData } = props;
  const isEditing = annotation.state === GraphNodeState.editing;
  const { theme } = useTheme();
  const rectWidth = getRectWidth<ICanvasAnnotation>(config, annotation);
  const rectHeight = getRectHeight<ICanvasAnnotation>(config, annotation);
  const containerStyles = getContainerStyles(annotation, theme, config);
  const textAreaStyles = {
    maxWidth: rectWidth - 22,
    maxHeight: rectHeight - 22
  };

  const comment = annotation.data && annotation.data.comment;

  return (
    <foreignObject
      x={annotation.x}
      y={annotation.y}
      height={rectHeight}
      width={rectWidth}
    >
      <div style={containerStyles}>
        {isEditing && (
          <textarea
            value={comment}
            placeholder="Input your comment here"
            onChange={getChangeHandler(annotation, setData)}
            style={textAreaStyles}
            rows={5}
            autoFocus={true}
          />
        )}
        {!isEditing && annotation.data && annotation.data.comment && (
          <div style={{ fontSize: 12, userSelect: "none" }}>{comment}</div>
        )}
      </div>
    </foreignObject>
  );
};
