import * as React from "react";
import { GraphNodeState, ICanvasNode, TSetData } from "../components";
import { INodeDrawArgs, IRectConfig, ITheme } from "../contexts";
import { updateNode } from "../utils";
import { getRectHeight, getRectWidth } from "../utils/layout";

const getChangeHandler = (node: ICanvasNode, setData: TSetData) => (
  e: React.ChangeEvent<HTMLInputElement>
) => {
  updateNode(
    node.id,
    { data: { ...node.data, comment: e.target.value } },
    setData
  );
};

export const rect: IRectConfig<ICanvasNode> = {
  getMinHeight(): number {
    return 150;
  },
  getMinWidth(): number {
    return 150;
  },
  getStyle(node: ICanvasNode, theme: ITheme): Partial<React.CSSProperties> {
    if (
      node.state === GraphNodeState.selected ||
      node.state === GraphNodeState.activated
    ) {
      return {
        fill: theme.nodeActivateFill,
        stroke: theme.nodeActivateStroke
      };
    }

    return {
      fill: theme.nodeFill,
      fillOpacity: 0.1,
      stroke: theme.nodeStroke,
      borderRadius: "5"
    };
  },
  render(args: INodeDrawArgs): React.ReactNode {
    const node = args.model;
    const graphProps = args.graphProps;
    const width = getRectWidth<ICanvasNode>(rect, node);
    const height = getRectHeight<ICanvasNode>(rect, node);
    const style = rect.getStyle ? rect.getStyle(node, args.theme) : {};
    const textY = node.y + height / 3;

    if (!graphProps.isNodeDraggingDisabled) {
      style.cursor = "move";
    }

    const comment = node.data && node.data.comment ? node.data.comment : "";

    return (
      <g key={node.id}>
        <rect
          width={width}
          height={height}
          x={node.x}
          y={node.y}
          style={style}
          rx={style.borderRadius}
        />
        <text x={node.x} y={textY} fontSize={12}>
          {node.name}
        </text>
        {node.data &&
          node.data.comment &&
          node.state !== GraphNodeState.editing && (
            <text
              x={node.x}
              y={textY + 20}
              fontSize={12}
              className={`comment-${node.id}`}
            >
              {node.data.comment}
            </text>
          )}
        {node.state === GraphNodeState.editing && (
          <foreignObject
            x={node.x}
            y={textY}
            height={height / 2.5}
            width={width - 5}
          >
            <input
              value={comment}
              placeholder="Input your comment here"
              onChange={getChangeHandler(node, args.setData.bind(null))}
            />
          </foreignObject>
        )}
      </g>
    );
  }
};
