import * as React from "react";
import { v4 as uuid } from "uuid";
import { MouseEventButton } from "../common/constants";
import {
  GraphAnnotationEvent,
  GraphCanvasEvent,
  GraphEdgeEvent,
  GraphNodeEvent,
  GraphPortEvent
} from "../common/GraphEvent.constant";
import {
  GraphConfigContext,
  IGraphConfig,
  IPropsAPIContext,
  MenuType,
  PanelContext,
  PropsAPIContext
} from "../contexts";
import {
  useConnectingLine,
  useContextMenu,
  useKeyboardHandlers,
  useTheme,
  useWheelHandlers
} from "../hooks";
import {
  addEdge,
  getClientByPoint,
  getEntityById,
  getNearestConnectablePort,
  getPointByClient,
  getPortPositionByPortId,
  getSelectedEntityIds,
  isConnectable,
  isSupported,
  moveNode,
  normalizePortsConnectState,
  resetPortsState,
  updateEntity,
  updatePort
} from "../utils";
import { graphController, GraphState } from "../utils/graphController";
import { initGraphStyle } from "../utils/initGraphStyle";
import { throttle } from "../utils/throttle";
import { dataStore } from "../utils/undoRedo";
import { isEdgeInViewPort, isNodeInViewPort } from "../utils/virtualization";
import {
  IZoomPanSettings,
  pan,
  scrollIntoView,
  TSetZoomPanSettings,
  zoomToFit
} from "../utils/zoomAndPan";
import { ConnectingLine } from "./ConnectingLine";
import {
  GraphCanvasEventProps,
  GraphEdgeEventProps,
  GraphEdgeState,
  GraphNodeEventProps,
  GraphNodeState,
  GraphPortEventProps,
  GraphPortState,
  ICanvasAnnotation,
  ICanvasData,
  ICanvasEdge,
  ICanvasNode,
  ICanvasPort,
  TSetData
} from "./Graph.interface";
import { GraphAnnotation } from "./GraphAnnotation";
import { GraphEdge } from "./GraphEdge";
import { GraphNode } from "./GraphNode";
import { GraphNodeControlPoints } from "./GraphNodeControlPoints";
import { GraphOneNodePorts } from "./GraphOneNodePorts";

import { GraphItemPanel } from "./ItemPanel";

import { SidePanel } from "./SidePanel";

interface ISelectBoxPosition {
  startX: number;
  startY: number;
  width: number;
  height: number;
}

export interface IGraphProps
  extends Partial<GraphNodeEventProps>,
    Partial<GraphEdgeEventProps>,
    Partial<GraphPortEventProps>,
    Partial<GraphCanvasEventProps> {
  style?: React.CSSProperties;
  data: ICanvasData;
  setData?: TSetData;
  zoomPanSettings?: IZoomPanSettings;
  setPanZoomPanSettings?: TSetZoomPanSettings;
  isNodeDraggingDisabled?: boolean;
  isNodeResizeDisabled?: boolean;
  isClickNodeToSelectDisabled?: boolean;
  isPropsAPIDisabled?: boolean;
  isPanDisabled?: boolean;
  isMultiSelectDisabled?: boolean;
  isConnectDisabled?: boolean;
  isSidePanelDisabled?: boolean;
  isAutoFitDisabled?: boolean;
  isKeyboardEventsDisabled?: boolean;
  isCanvasFocusDisabled?: boolean;
  isAddingNodeDisabled?: boolean;
  isScrollDisabled?: boolean;
  svgRef?: React.RefObject<SVGSVGElement>;
}

let lastClientX = 0;
let lastClientY = 0;
let draggingNodeId: string;
let isCanvasFocused = false;

let curConnectingNode: ICanvasNode | undefined;
let curConnectingPort: ICanvasPort | undefined;
let curConnectingTargetNode: ICanvasNode | undefined;
let curConnectingTargetPort: ICanvasPort | undefined;

// tslint:disable-next-line: cyclomatic-complexity
export const Graph: React.FunctionComponent<IGraphProps> = props => {
  React.useEffect(() => {
    dataStore.record(props.data);
  }, []);

  if (!isSupported()) {
    return (
      <p>
        Your browser is not supported. Current supported browsers: Google
        Chrome, Microsoft Edge based on Chromium
      </p>
    );
  }

  let [data, setData] = React.useState<ICanvasData>(props.data);

  if (props.data) {
    data = props.data;
    setData = props.setData
      ? props.setData
      : () => {
          // default setData
        };
  }

  let [zoomPanSettings, setZoomPanSettings] = React.useState<IZoomPanSettings>(
    props.zoomPanSettings
      ? props.zoomPanSettings
      : {
          transformMatrix: [1, 0, 0, 1, 0, 0]
        }
  );

  if (props.setPanZoomPanSettings) {
    zoomPanSettings = props.zoomPanSettings
      ? props.zoomPanSettings
      : {
          transformMatrix: [1, 0, 0, 1, 0, 0]
        };
    setZoomPanSettings = props.setPanZoomPanSettings;
  }

  const { propsAPI } = React.useContext<IPropsAPIContext>(PropsAPIContext);
  const { theme } = useTheme();
  const svgRef = props.svgRef
    ? props.svgRef
    : React.useRef<SVGSVGElement>(null);
  const graphConfig = React.useContext<IGraphConfig>(GraphConfigContext);

  if (!props.isPropsAPIDisabled) {
    propsAPI.init(
      setData,
      () => data,
      svgRef,
      zoomPanSettings,
      setZoomPanSettings,
      graphConfig,
      React.useContext(PanelContext)
    );
  }

  const [isSpaceHeld, setIsSpaceHeld] = React.useState<boolean>(false);

  const [connectingState, setConnectingState] = useConnectingLine();
  const [tempConnectLine, setTempConnectLine] = useConnectingLine();

  const [selectBoxPosition, setSelectBoxPosition] = React.useState<
    ISelectBoxPosition
  >({
    startX: 0,
    startY: 0,
    height: 0,
    width: 0
  });

  const [
    contextMenuEle,
    setCurContextMenuType,
    isVisibleContextMenu,
    setIsVisibleContextMenu,
    contextMenuRef,
    contextMenuPosition
  ] = useContextMenu(false, lastClientX, lastClientY);

  const [visibleAddingNodeSvg, setVisibleAddingNodeSvg] = React.useState(false);

  const graphId = `graph-${uuid()}`;
  const { nodes, edges } = normalizePortsConnectState(data);
  const isNodesDraggable = !props.isNodeDraggingDisabled;
  const isClickNodeToSelectEnabled = !props.isClickNodeToSelectDisabled;
  const isNodeResizable = !props.isNodeResizeDisabled;

  if (!props.isKeyboardEventsDisabled) {
    useKeyboardHandlers({
      setIsSpaceHeld,
      getGraphConfig: () => graphConfig,
      // getData: () => data,
      setData,
      getIsCanvasFocused: () => isCanvasFocused
    });
  }

  if (!props.isAutoFitDisabled) {
    React.useEffect(() => {
      zoomToFit(data, setZoomPanSettings, svgRef);
    }, [[...data.nodes.map(n => n.x), ...data.nodes.map(n => n.y)].join(",")]);
  }

  React.useEffect(() => {
    return () => {
      if (selectBoxPosition.width !== 0 && selectBoxPosition.height !== 0) {
        const selectAreaMinX = Math.min(
          selectBoxPosition.startX,
          selectBoxPosition.startX + selectBoxPosition.width
        );
        const selectAreaMaxX = Math.max(
          selectBoxPosition.startX,
          selectBoxPosition.startX + selectBoxPosition.width
        );
        const selectAreaMinY = Math.min(
          selectBoxPosition.startY,
          selectBoxPosition.startY + selectBoxPosition.height
        );
        const selectAreaMaxY = Math.max(
          selectBoxPosition.startY,
          selectBoxPosition.startY + selectBoxPosition.height
        );

        const nextNodes = nodes.map(n => {
          const { x: realX, y: realY } = getClientByPoint(
            n.x,
            n.y,
            zoomPanSettings.transformMatrix
          );

          if (
            realX > selectAreaMinX &&
            realX < selectAreaMaxX &&
            realY > selectAreaMinY &&
            realY < selectAreaMaxY
          ) {
            return {
              ...n,
              state: GraphNodeState.selected
            };
          }

          return { ...n, state: GraphNodeState.default };
        });

        setData(prevData => ({
          ...prevData,
          nodes: nextNodes
        }));
      }
    };
  }, [
    selectBoxPosition.height,
    selectBoxPosition.width,
    selectBoxPosition.startX,
    selectBoxPosition.startY
  ]);

  const unSelectAll = (exceptNodeId?: string) => {
    setData(prevState => {
      const nextNodes = prevState.nodes.map(n => {
        if (exceptNodeId && n.id === exceptNodeId) {
          return { ...n, state: GraphNodeState.selected };
        }

        return { ...n, state: GraphNodeState.default };
      });
      const nextAnnotations = prevState.annotations?.map(n => {
        if (exceptNodeId && n.id === exceptNodeId) {
          return { ...n, state: GraphNodeState.selected };
        }

        return { ...n, state: GraphNodeState.default };
      });

      const nextEdges = prevState.edges.map(e => {
        return { ...e, state: GraphEdgeState.default };
      });

      return resetPortsState({
        nodes: nextNodes,
        edges: nextEdges,
        annotations: nextAnnotations
      });
    });
  };

  //#region container handler
  const onContainerMouseDown = (
    evt: React.MouseEvent<SVGSVGElement, MouseEvent>
  ) => {
    evt.stopPropagation();
    evt.preventDefault();

    const svgRect = (evt.target as SVGSVGElement).getBoundingClientRect();

    lastClientX = evt.clientX;
    lastClientY = evt.clientY;

    const currentX = evt.clientX - svgRect.left;
    const currentY = evt.clientY - svgRect.top;

    if (isClickNodeToSelectEnabled) {
      unSelectAll();
    }

    window.addEventListener("mousemove", handleDrag);
    window.addEventListener("mouseup", handleDragEnd);

    if (isSpaceHeld && !props.isPanDisabled) {
      graphController.setGraphState(GraphState.panning);
      if (props.onCanvasDragStart) {
        props.onCanvasDragStart(evt);
      }
    } else if (!props.isMultiSelectDisabled) {
      graphController.setGraphState(GraphState.multiSelect);

      setSelectBoxPosition(prevState => ({
        ...prevState,
        startX: currentX,
        startY: currentY,
        width: 0,
        height: 0
      }));
    }
    setIsVisibleContextMenu(false);
  };

  const onContainerMouseEnter = () => {
    // TODO: Use real focus/blur to update this flag after a11y keyboard interaction completed
    if (!props.isCanvasFocusDisabled) {
      isCanvasFocused = true;
    }

    if (
      graphController.getGraphState() === GraphState.addingNode &&
      !props.isAddingNodeDisabled
    ) {
      setVisibleAddingNodeSvg(true);
      window.addEventListener("mouseup", handleDragEnd);
    }
  };

  const onContainerMouseLeave = () => {
    if (!props.isCanvasFocusDisabled) {
      isCanvasFocused = false;
    }
  };

  const canvasEventHandler = (eventName: GraphCanvasEvent) => (
    evt: React.MouseEvent<SVGSVGElement, MouseEvent>
  ) => {
    switch (eventName) {
      case GraphCanvasEvent.onCanvasMouseDown:
        onContainerMouseDown(evt);
        break;
      case GraphCanvasEvent.onCanvasMouseEnter:
        onContainerMouseEnter();
        break;
      case GraphCanvasEvent.onCanvasMouseLeave:
        onContainerMouseLeave();
        break;
      case GraphCanvasEvent.onCanvasContextMenu:
        onContextMenu(evt, MenuType.Canvas);
        break;
      case GraphCanvasEvent.onCanvasDoubleClick:
        propsAPI.addAnnotation({
          id: uuid(),
          x: evt.clientX,
          y: evt.clientY
        });
        break;
      default:
    }

    const userEventHandler = props[eventName];
    if (userEventHandler) {
      userEventHandler(evt);
    }
  };
  //#endregion container handler

  const handleDrag = throttle((evt: MouseEvent): void => {
    const clientDX = evt.clientX - lastClientX;
    const clientDY = evt.clientY - lastClientY;

    const { x: dx, y: dy } = getPointByClient(
      clientDX,
      clientDY,
      zoomPanSettings.transformMatrix
    );

    lastClientX = evt.clientX;
    lastClientY = evt.clientY;

    switch (graphController.getGraphState()) {
      case GraphState.dragging:
        moveNode(draggingNodeId, dx, dy, setData);

        if (props.onNodeDrag) {
          props.onNodeDrag(evt, getEntityById(draggingNodeId, data.nodes));
        }
        break;

      case GraphState.panning:
        if (props.onCanvasDrag) {
          props.onCanvasDrag(evt);
        }
        pan(dx, dy, setZoomPanSettings);
        break;

      case GraphState.multiSelect:
        setSelectBoxPosition(prevState => {
          return {
            ...prevState,
            width: prevState.width + clientDX,
            height: prevState.height + clientDY
          };
        });
        break;

      case GraphState.connecting:
        setConnectingState(prevState => {
          return {
            ...prevState,
            x2: prevState.x2 + clientDX,
            y2: prevState.y2 + clientDY
          };
        });
        break;
      default:
    }
  }, 20);

  const handleDragEnd = (evt: MouseEvent) => {
    switch (graphController.getGraphState()) {
      case GraphState.dragging:
        if (svgRef.current) {
          const containerEl = svgRef.current.closest<HTMLDivElement>(
            ".react-dag-editor-container"
          );

          if (containerEl) {
            scrollIntoView(
              evt.clientX,
              evt.clientY,
              svgRef,
              setZoomPanSettings
            );
          }
          if (props.onNodeDragEnd) {
            props.onNodeDragEnd(evt, getEntityById(draggingNodeId, data.nodes));
          }
        }
        break;
      case GraphState.connecting:
        connectingPortsEnd();
        break;
      case GraphState.panning:
        if (props.onCanvasDragEnd) {
          props.onCanvasDragEnd(evt);
        }
        break;
      default:
    }

    draggingNodeId = "";
    graphController.setGraphState(GraphState.default);
    setVisibleAddingNodeSvg(false);

    setSelectBoxPosition(prevState => ({
      ...prevState,
      startX: 0,
      startY: 0,
      height: 0,
      width: 0
    }));

    setConnectingState({
      x1: 0,
      y1: 0,
      x2: 0,
      y2: 0,
      visible: false
    });

    setTempConnectLine({
      x1: 0,
      y1: 0,
      x2: 0,
      y2: 0,
      visible: false
    });

    window.removeEventListener("mousemove", handleDrag);
    window.removeEventListener("mouseup", handleDragEnd);
  };

  const connectingPortsEnd = () => {
    if (curConnectingNode && curConnectingPort) {
      setData(resetPortsState);
      if (curConnectingTargetNode && curConnectingTargetPort) {
        addEdge(
          {
            source: curConnectingNode.id,
            sourcePortId: curConnectingPort.id,
            target: curConnectingTargetNode.id,
            targetPortId: curConnectingTargetPort.id,
            id: uuid()
          },
          setData
        );
        curConnectingTargetNode = undefined;
        curConnectingTargetPort = undefined;
      }
    }
  };

  const autoAttachPort = () => {
    if (
      curConnectingNode &&
      curConnectingPort &&
      curConnectingTargetNode &&
      curConnectingTargetPort
    ) {
      const targetPortPos = getPortPositionByPortId(
        curConnectingTargetNode,
        curConnectingTargetPort.id,
        graphConfig,
        props,
        setData,
        zoomPanSettings.transformMatrix
      );
      const curConnectingPortPos = getPortPositionByPortId(
        curConnectingNode,
        curConnectingPort.id,
        graphConfig,
        props,
        setData,
        zoomPanSettings.transformMatrix
      );

      if (targetPortPos && curConnectingPortPos) {
        setTempConnectLine({
          x1: curConnectingPortPos.x,
          y1: curConnectingPortPos.y,
          x2: targetPortPos.x,
          y2: targetPortPos.y,
          visible: true
        });
        setConnectingState(prevState => {
          return {
            ...prevState,
            visible: false
          };
        });
      }
    }
  };

  const resetToConnectingLine = () => {
    curConnectingTargetNode = undefined;
    curConnectingTargetPort = undefined;
    setTempConnectLine(prevState => {
      return {
        ...prevState,
        visible: false
      };
    });
    setConnectingState(prevState => {
      return {
        ...prevState,
        visible: true
      };
    });
  };

  //#region node handler
  const onNodeMouseDown = (evt: React.MouseEvent, node: ICanvasNode) => {
    evt.stopPropagation();
    selectionEffect<ICanvasNode>(evt, node, "nodes");

    if (isNodesDraggable) {
      graphController.setGraphState(GraphState.dragging);

      draggingNodeId = node.id;

      if (props.onNodeDragStart) {
        props.onNodeDragStart(evt, getEntityById(draggingNodeId, data.nodes));
      }

      window.addEventListener("mousemove", handleDrag);
      window.addEventListener("mouseup", handleDragEnd);
    }
  };

  const onNodeMouseEnter = (evt: React.MouseEvent, node: ICanvasNode) => {
    hoverEffect<ICanvasNode>(node, data.nodes, "nodes");

    if (graphController.getGraphState() === GraphState.connecting) {
      curConnectingTargetNode = node;
    }
  };

  const onNodeMouseLeave = (evt: React.MouseEvent, node: ICanvasNode) => {
    unHoverEffect(node, data.nodes, "nodes");

    if (graphController.getGraphState() === GraphState.connecting) {
      resetToConnectingLine();
    }
  };

  const onNodeMouseMove = (evt: React.MouseEvent, node: ICanvasNode) => {
    if (graphController.getGraphState() === GraphState.connecting) {
      const clientDX = evt.clientX - lastClientX;
      const clientDY = evt.clientY - lastClientY;

      setConnectingState(state => {
        if (curConnectingTargetNode) {
          curConnectingTargetPort = getNearestConnectablePort(
            curConnectingTargetNode,
            state.x2 + clientDX,
            state.y2 + clientDY,
            graphConfig,
            props,
            setData,
            zoomPanSettings.transformMatrix
          );
          autoAttachPort();
        }

        return { ...state };
      });
    }
  };

  // handle all node events
  const nodeEventHandler = (node: ICanvasNode, eventName: GraphNodeEvent) => (
    evt: React.MouseEvent
  ) => {
    switch (eventName) {
      case GraphNodeEvent.onNodeMouseDown:
        onNodeMouseDown(evt, node);
        break;
      case GraphNodeEvent.onNodeMouseEnter:
        onNodeMouseEnter(evt, node);
        break;
      case GraphNodeEvent.onNodeMouseLeave:
        onNodeMouseLeave(evt, node);
        break;
      case GraphNodeEvent.onNodeContextMenu:
        getSelectedEntityIds<ICanvasNode>(nodes).length > 1
          ? onContextMenu(evt, MenuType.Multi)
          : onContextMenu(evt, MenuType.Node);
        break;
      case GraphNodeEvent.onNodeMouseMove:
        onNodeMouseMove(evt, node);
        break;
      default:
    }

    const userEventHandler = props[eventName];
    if (userEventHandler) {
      userEventHandler(evt, node);
    }
  };
  //#endregion node handler

  // tslint:disable: no-any
  function hoverEffect<T extends { id: string; state?: GraphNodeState }>(
    entity: T,
    entities: T[],
    entityType: keyof ICanvasData
  ): void {
    const graphState = graphController.getGraphState();
    if (
      graphState !== GraphState.default &&
      graphState !== GraphState.connecting
    ) {
      return;
    }

    const curNode = getEntityById(entity.id, entities);
    if (
      curNode &&
      (curNode.state === GraphNodeState.default || !curNode.state)
    ) {
      const patch = { state: GraphNodeState.activated };
      updateEntity<T>(entity.id, entityType, patch as Partial<T>, setData);
    }
  }

  function unHoverEffect<T extends { id: string; state?: GraphNodeState }>(
    entity: T,
    entities: T[],
    entityType: keyof ICanvasData
  ): void {
    const graphState = graphController.getGraphState();
    if (
      graphState !== GraphState.default &&
      graphState !== GraphState.connecting
    ) {
      return;
    }

    const curNode = getEntityById(entity.id, entities);

    if (curNode && curNode.state === GraphNodeState.activated) {
      const patch = { state: GraphNodeState.default };
      updateEntity<T>(entity.id, entityType, patch as Partial<T>, setData);
    }
  }

  function selectionEffect<T extends { id: string; state?: GraphNodeState }>(
    evt: React.MouseEvent,
    entity: T,
    entityType: keyof ICanvasData
  ): void {
    evt.stopPropagation();

    lastClientX = evt.clientX;
    lastClientY = evt.clientY;

    if (!isClickNodeToSelectEnabled) {
      return;
    }
    if (entity.state === GraphNodeState.editing) {
      return;
    }

    const entityTypeStr = entityType as string;
    if (!evt.ctrlKey) {
      const selectedEntityIds = getSelectedEntityIds<T>(data[entityTypeStr]);
      const isMultiSelectContextMenu =
        evt.button === MouseEventButton.Right &&
        selectedEntityIds.length > 1 &&
        selectedEntityIds.indexOf(entity.id) >= 0;

      if (!isMultiSelectContextMenu) {
        unSelectAll(entity.id);
      }
    } else {
      const patch = { state: GraphNodeState.selected };
      updateEntity<T>(entity.id, entityType, patch as Partial<T>, setData);
    }

    setIsVisibleContextMenu(false);
  }

  //#region edge handler
  // handle all edge events
  const edgeEventHandler = (edge: ICanvasEdge, eventName: GraphEdgeEvent) => (
    evt: React.MouseEvent<SVGGElement>
  ) => {
    switch (eventName) {
      case GraphEdgeEvent.onEdgeContextMenu:
        onContextMenu(evt, MenuType.Edge);
        break;
      default:
    }

    const userEventHandler = props[eventName];
    if (userEventHandler) {
      userEventHandler(evt, edge);
    }
  };
  //#endregion edge handler

  //#region ports handler
  const onPortMouseDown = (
    evt: React.MouseEvent,
    node: ICanvasNode,
    port: ICanvasPort
  ) => {
    if (props.isConnectDisabled) {
      return;
    }

    evt.stopPropagation();
    curConnectingNode = node;
    curConnectingPort = port;
    lastClientX = evt.clientX;
    lastClientY = evt.clientY;

    const portPos = getPortPositionByPortId(
      node,
      port.id,
      graphConfig,
      props,
      setData,
      zoomPanSettings.transformMatrix
    );

    setConnectingState(prevState => {
      return {
        ...prevState,
        x1: portPos.x,
        y1: portPos.y,
        x2: portPos.x + 8,
        y2: portPos.y + 14,
        visible: true
      };
    });
    graphController.setGraphState(GraphState.connecting);
    updatePort(node.id, port.id, { state: GraphPortState.connecting }, setData);
    window.addEventListener("mousemove", handleDrag);
    window.addEventListener("mouseup", handleDragEnd);

    setIsVisibleContextMenu(false);
  };

  const onPortMouseEnter = (
    evt: React.MouseEvent,
    node: ICanvasNode,
    port: ICanvasPort
  ) => {
    if (graphController.getGraphState() === GraphState.connecting) {
      curConnectingTargetNode = node;
      setData(state => {
        if (isConnectable(graphConfig, port, state)) {
          curConnectingTargetPort = port;
          autoAttachPort();
        }
        return { ...state };
      });
    }
  };
  const onPortMouseLeave = () => {
    if (graphController.getGraphState() === GraphState.connecting) {
      resetToConnectingLine();
    }
  };

  const portEventHandler = (
    node: ICanvasNode,
    port: ICanvasPort,
    eventName: GraphPortEvent
  ) => (evt: React.MouseEvent<SVGGElement>) => {
    switch (eventName) {
      case GraphPortEvent.onPortMouseEnter:
        onPortMouseEnter(evt, node, port);
        break;
      case GraphPortEvent.onPortMouseLeave:
        onPortMouseLeave();
        break;
      case GraphPortEvent.onPortMouseDown:
        onPortMouseDown(evt, node, port);
        break;
      case GraphPortEvent.onPortContextMenu:
        onContextMenu(evt, MenuType.Port);
        break;
      default:
    }

    const userEventHandler = props[eventName];
    if (userEventHandler) {
      userEventHandler(evt, node, port);
    }
  };
  //#endregion ports handler

  const onContextMenu = (evt: React.MouseEvent, menuType: MenuType) => {
    evt.preventDefault();
    evt.stopPropagation();
    setCurContextMenuType(menuType);
    setIsVisibleContextMenu(true);
  };

  //#region annotation
  const annotationEventHandler = (
    annotation: ICanvasAnnotation,
    eventName: GraphAnnotationEvent
  ) => (evt: React.MouseEvent) => {
    switch (eventName) {
      case GraphAnnotationEvent.onAnnotationMouseDown:
        selectionEffect<ICanvasAnnotation>(evt, annotation, "annotations");
        break;
      case GraphAnnotationEvent.onAnnotationMouseEnter:
        hoverEffect<ICanvasAnnotation>(
          annotation,
          data.annotations || [],
          "annotations"
        );
        break;
      case GraphAnnotationEvent.onAnnotationMouseLeave:
        unHoverEffect(annotation, data.annotations || [], "annotations");
        break;
      case GraphAnnotationEvent.onAnnotationContextMenu:
        onContextMenu(evt, MenuType.Annotation);
        break;
      default:
    }

    const userEventHandler = props[eventName];
    if (userEventHandler) {
      userEventHandler(evt, annotationEventHandler);
    }
  };

  const renderAnnotation = (annotation: ICanvasAnnotation) => (
    <g key={annotation.id}>
      <GraphAnnotation
        annotation={annotation}
        setData={setData}
        graphProps={props}
        onAnnotationEvent={annotationEventHandler}
      />
    </g>
  );
  //#endregion annotation

  const renderNode = (node: ICanvasNode) => (
    <g key={node.id}>
      <GraphNode
        node={node}
        setData={setData}
        graphProps={props}
        onNodeEvent={nodeEventHandler}
      />
      <GraphOneNodePorts
        node={node}
        data={data}
        setData={setData}
        graphProps={props}
        onPortEvent={portEventHandler}
      />
    </g>
  );

  const selectBoxD = `m${selectBoxPosition.startX} ${
    selectBoxPosition.startY
  } v ${selectBoxPosition.height} h ${
    selectBoxPosition.width
  } v${-selectBoxPosition.height} h ${-selectBoxPosition.width}`;

  const svgStyle: React.CSSProperties = {
    ...initGraphStyle(),
    cursor: isSpaceHeld ? "grab" : "initial"
  };
  const transform = `matrix(${zoomPanSettings.transformMatrix.join(" ")})`;

  if (!props.isScrollDisabled) {
    useWheelHandlers({
      getIsCanvasFocused: () => isCanvasFocused,
      setZoomPanSettings,
      svgRef
    });
  }

  const containerStyles: React.CSSProperties = {
    position: "relative",
    width: "100%",
    height: "100%",
    overflow: "hidden",
    background: theme.canvasBackground,
    ...props.style
  };

  const visibleNodes = nodes.filter(node =>
    isNodeInViewPort(node, zoomPanSettings.transformMatrix)
  );

  const visibleEdges = edges.filter(edge =>
    isEdgeInViewPort(edge, data, zoomPanSettings.transformMatrix)
  );

  const onContextMenuClick = (evt: React.MouseEvent) => {
    evt.preventDefault();
    setIsVisibleContextMenu(false);
  };
  const onContextMenuMouseEnter = (evt: React.MouseEvent) => {
    isCanvasFocused = true;
  };

  const renderContextMenu = (): React.ReactNode => {
    const { left, top, right, bottom } = contextMenuPosition;
    return (
      isVisibleContextMenu && (
        <div
          ref={contextMenuRef}
          onClick={onContextMenuClick}
          onMouseEnter={onContextMenuMouseEnter}
          onContextMenu={onContextMenuClick}
          role="button"
          style={{
            position: "fixed",
            left: left || undefined,
            top: top || undefined,
            right: left ? undefined : right,
            bottom: top ? undefined : bottom
          }}
        >
          {contextMenuEle}
        </div>
      )
    );
  };

  return (
    <div
      style={containerStyles}
      id={graphId}
      className="react-dag-editor-container"
    >
      <svg
        onClick={canvasEventHandler(GraphCanvasEvent.onCanvasClick)}
        onDoubleClick={canvasEventHandler(GraphCanvasEvent.onCanvasDoubleClick)}
        onMouseEnter={canvasEventHandler(GraphCanvasEvent.onCanvasMouseEnter)}
        onMouseLeave={canvasEventHandler(GraphCanvasEvent.onCanvasMouseLeave)}
        onMouseDown={canvasEventHandler(GraphCanvasEvent.onCanvasMouseDown)}
        onMouseUp={canvasEventHandler(GraphCanvasEvent.onCanvasMouseUp)}
        onContextMenu={canvasEventHandler(GraphCanvasEvent.onCanvasContextMenu)}
        onMouseMove={canvasEventHandler(GraphCanvasEvent.onCanvasMouseMove)}
        onMouseOver={canvasEventHandler(GraphCanvasEvent.onCanvasMouseOver)}
        onMouseOut={canvasEventHandler(GraphCanvasEvent.onCanvasMouseOut)}
        onFocus={undefined}
        onBlur={undefined}
        style={svgStyle}
        preserveAspectRatio="xMidYMid meet"
        ref={svgRef}
      >
        <g transform={transform}>
          {data.annotations?.map(renderAnnotation)}
          {visibleEdges.map(edge => (
            <GraphEdge
              key={edge.id}
              edge={edge}
              nodes={nodes}
              graphProps={props}
              setData={setData}
              onEdgeEvent={edgeEventHandler}
            />
          ))}
          {visibleNodes.map(renderNode)}
          {isNodeResizable &&
            visibleNodes.map(node => (
              <GraphNodeControlPoints
                node={node}
                key={node.id}
                zoomPanSettings={zoomPanSettings}
              />
            ))}
        </g>
        {!props.isMultiSelectDisabled && (
          <path fill="none" stroke={theme.defaultColor} d={selectBoxD} />
        )}
        <ConnectingLine
          connectingLine={connectingState}
          autoAttachLine={tempConnectLine}
        />
      </svg>
      <GraphItemPanel
        visible={visibleAddingNodeSvg}
        setData={setData}
        graphProps={props}
        graphSvgRef={svgRef}
        zoomPanSettings={zoomPanSettings}
      />
      {renderContextMenu()}
      {!props.isSidePanelDisabled && <SidePanel svgRef={svgRef} />}
    </div>
  );
  // TODO: Fix this rule
  // tslint:disable-next-line: max-file-line-count
};
