import * as React from "react";
import { GraphAnnotationEvent } from "../common/GraphEvent.constant";
import { GraphConfigContext, IGraphConfig } from "../contexts";
import { useTheme } from "../hooks";
import { updateAnnotation } from "../utils";
import { IGraphProps } from "./Graph";
import { GraphNodeState, ICanvasAnnotation, TSetData } from "./Graph.interface";

interface IProps {
  annotation: ICanvasAnnotation;
  setData: TSetData;
  graphProps: IGraphProps;
  onAnnotationEvent(
    annotation: ICanvasAnnotation,
    eventName: GraphAnnotationEvent
  ): React.EventHandler<React.MouseEvent<SVGGElement>>;
}

const GraphAnnotation: React.FunctionComponent<IProps> = props => {
  const { annotation: anno, graphProps, onAnnotationEvent } = props;
  const setData = props.setData.bind(null);
  const graphConfig = React.useContext<IGraphConfig>(GraphConfigContext);
  const shape = anno.shape ? anno.shape : "annotation";
  const annotationConfig = graphConfig.getAnnotationConfigByName(shape);
  const { theme } = useTheme();

  if (!annotationConfig) {
    throw new Error(`invalid shape in annotation ${anno.id}`);
  }
  if (!annotationConfig.render) {
    throw new Error(`Missing "render" method in annotation config ${shape}`);
  }

  const {
    onAnnotationClick,
    onAnnotationDoubleClick,
    onAnnotationMouseDown,
    onAnnotationMouseUp,
    onAnnotationMouseEnter,
    onAnnotationMouseLeave,
    onAnnotationContextMenu,
    onAnnotationMouseMove,
    onAnnotationMouseOver,
    onAnnotationMouseOut
  } = GraphAnnotationEvent;

  // add comment to annotation
  const onDoubleClick = (evt: React.MouseEvent<SVGGElement>) => {
    evt.stopPropagation();

    updateAnnotation(
      anno.id,
      {
        state: GraphNodeState.editing
      },
      setData
    );
    onAnnotationEvent(anno, onAnnotationDoubleClick)(evt);
  };

  return (
    <g
      key={anno.id}
      id={`annotation-container-${anno.id}`}
      className="annotation-box-container"
      onClick={onAnnotationEvent(anno, onAnnotationClick)}
      onDoubleClick={onDoubleClick}
      onMouseDown={onAnnotationEvent(anno, onAnnotationMouseDown)}
      onMouseUp={onAnnotationEvent(anno, onAnnotationMouseUp)}
      onMouseEnter={onAnnotationEvent(anno, onAnnotationMouseEnter)}
      onMouseLeave={onAnnotationEvent(anno, onAnnotationMouseLeave)}
      onContextMenu={onAnnotationEvent(anno, onAnnotationContextMenu)}
      onMouseMove={onAnnotationEvent(anno, onAnnotationMouseMove)}
      onMouseOver={onAnnotationEvent(anno, onAnnotationMouseOver)}
      onMouseOut={onAnnotationEvent(anno, onAnnotationMouseOut)}
      onFocus={undefined}
      onBlur={undefined}
    >
      {annotationConfig.render({
        model: anno,
        graphProps,
        setData: props.setData,
        theme
      })}
    </g>
  );
};

const GraphAnnoMemo = React.memo(GraphAnnotation, (prevProps, nextProps) => {
  const prevData = prevProps.annotation;
  const nextData = nextProps.annotation;

  return ["id", "name", "state", "data", "x", "y"].every(
    key => prevData[key] === nextData[key]
  );
});

export { GraphAnnoMemo as GraphAnnotation };
