import * as React from "react";
import { GraphEdgeEvent } from "../common/GraphEvent.constant";
import { GraphConfigContext, IGraphConfig } from "../contexts";
import { useTheme } from "../hooks";
import { getPortPositionByPortId, updateEdge } from "../utils";
import { IGraphProps } from "./Graph";
import {
  GraphEdgeState,
  ICanvasEdge,
  ICanvasNode,
  TSetData
} from "./Graph.interface";

interface IProps {
  edge: ICanvasEdge;
  nodes: ICanvasNode[];
  graphProps: IGraphProps;
  setData: TSetData;
  onEdgeEvent(
    edge: ICanvasEdge,
    eventName: GraphEdgeEvent
  ): React.EventHandler<React.MouseEvent<SVGGElement>>;
}

export const GraphEdge: React.FunctionComponent<IProps> = props => {
  const { edge, nodes, graphProps, onEdgeEvent } = props;
  const setData = props.setData.bind(null);
  const sourceNode = nodes.find(node => node.id === edge.source);
  const graphConfig = React.useContext<IGraphConfig>(GraphConfigContext);

  if (!sourceNode) {
    throw new Error(`invalid source ${JSON.stringify(edge)}`);
  }

  const { theme } = useTheme();

  const sourcePosition = getPortPositionByPortId(
    sourceNode,
    edge.sourcePortId,
    graphConfig,
    graphProps,
    setData
  );

  const targetNode = nodes.find(node => node.id === edge.target);

  if (!targetNode) {
    throw new Error("invalid target");
  }

  const targetPosition = getPortPositionByPortId(
    targetNode,
    edge.targetPortId,
    graphConfig,
    graphProps,
    setData
  );

  const x1 = sourcePosition.x;
  const y1 = sourcePosition.y;
  const x2 = targetPosition.x;
  const y2 = targetPosition.y;

  const shape = edge.shape ? edge.shape : "line";
  const edgeConfig = graphConfig.getEdgeConfigByName(shape);

  if (!edgeConfig) {
    throw new Error(`invalid shape in edge ${edge.id}`);
  }

  if (!edgeConfig.render) {
    throw new Error(`Missing "render" method in edge config ${shape}`);
  }

  const onClick = (evt: React.MouseEvent<SVGGElement>) => {
    evt.stopPropagation();
    updateEdge(edge.id, { state: GraphEdgeState.selected }, setData);
    onEdgeEvent(edge, GraphEdgeEvent.onEdgeClick)(evt);
  };

  const onContextMenu = (evt: React.MouseEvent<SVGGElement>) => {
    updateEdge(edge.id, { state: GraphEdgeState.selected }, setData);
    onEdgeEvent(edge, GraphEdgeEvent.onEdgeContextMenu)(evt);
  };

  return (
    <g
      onClick={onClick}
      onDoubleClick={onEdgeEvent(edge, GraphEdgeEvent.onEdgeDoubleClick)}
      onMouseDown={onEdgeEvent(edge, GraphEdgeEvent.onEdgeMouseDown)}
      onMouseUp={onEdgeEvent(edge, GraphEdgeEvent.onEdgeMouseUp)}
      onMouseEnter={onEdgeEvent(edge, GraphEdgeEvent.onEdgeMouseEnter)}
      onMouseLeave={onEdgeEvent(edge, GraphEdgeEvent.onEdgeMouseLeave)}
      onContextMenu={onContextMenu}
      onMouseMove={onEdgeEvent(edge, GraphEdgeEvent.onEdgeMouseMove)}
      onMouseOver={onEdgeEvent(edge, GraphEdgeEvent.onEdgeMouseOver)}
      onMouseOut={onEdgeEvent(edge, GraphEdgeEvent.onEdgeMouseOut)}
      onFocus={undefined}
      onBlur={undefined}
    >
      {edgeConfig.render({
        model: edge,
        graphProps,
        x1,
        y1,
        x2,
        y2,
        setData,
        theme
      })}
    </g>
  );
};
