import * as React from "react";
import { GraphNodeEvent } from "../common/GraphEvent.constant";
import { GraphConfigContext, IGraphConfig } from "../contexts";
import { useTheme } from "../hooks";
import { getNodeConfig, updateNode } from "../utils";
import { IGraphProps } from "./Graph";
import { GraphNodeState, ICanvasNode, TSetData } from "./Graph.interface";

interface IProps {
  node: ICanvasNode;
  graphProps: IGraphProps;
  setData: TSetData;
  onNodeEvent?(
    node: ICanvasNode,
    eventName: GraphNodeEvent
  ): React.EventHandler<React.MouseEvent<SVGGElement>>;
}

const GraphNode: React.FunctionComponent<IProps> = props => {
  const { node, graphProps, onNodeEvent } = props;
  const setData = props.setData.bind(null);
  const graphConfig = React.useContext<IGraphConfig>(GraphConfigContext);
  const shape = node.shape ? node.shape : "rect";
  const nodeConfig = getNodeConfig(node, graphConfig);
  const { theme } = useTheme();

  if (!nodeConfig.render) {
    throw new Error(`Missing "render" method in node config ${shape}`);
  }

  // add comment to node
  const onDoubleClick = (evt: React.MouseEvent<SVGGElement>) => {
    evt.stopPropagation();

    updateNode(
      node.id,
      {
        state: GraphNodeState.editing
      },
      setData
    );
    nodeEvent(GraphNodeEvent.onNodeDoubleClick)(evt);
  };

  const nodeEvent = (eventName: GraphNodeEvent) => (
    evt: React.MouseEvent<SVGGElement>
  ) => {
    if (onNodeEvent) {
      onNodeEvent(node, eventName)(evt);
    }
  };

  return (
    <g
      key={node.id}
      id={`node-container-${node.id}`}
      onClick={nodeEvent(GraphNodeEvent.onNodeClick)}
      onDoubleClick={onDoubleClick}
      onMouseDown={nodeEvent(GraphNodeEvent.onNodeMouseDown)}
      onMouseUp={nodeEvent(GraphNodeEvent.onNodeMouseUp)}
      onMouseEnter={nodeEvent(GraphNodeEvent.onNodeMouseEnter)}
      onMouseLeave={nodeEvent(GraphNodeEvent.onNodeMouseLeave)}
      onContextMenu={nodeEvent(GraphNodeEvent.onNodeContextMenu)}
      onMouseMove={nodeEvent(GraphNodeEvent.onNodeMouseMove)}
      onMouseOver={nodeEvent(GraphNodeEvent.onNodeMouseOver)}
      onMouseOut={nodeEvent(GraphNodeEvent.onNodeMouseOut)}
      onFocus={undefined}
      onBlur={undefined}
    >
      <g className="node-box-container">
        {nodeConfig.render({
          model: node,
          graphProps,
          setData: props.setData,
          theme
        })}
      </g>
    </g>
  );
};

const GraphNodeMemo = React.memo(GraphNode, (prevProps, nextProps) => {
  const prevNode = prevProps.node;
  const nextNode = nextProps.node;

  return ["id", "name", "state", "data", "x", "y", "ports"].every(
    key => prevNode[key] === nextNode[key]
  );
});

export { GraphNodeMemo as GraphNode };
