import * as React from "react";
import {
  GraphConfigContext,
  IGraphConfig,
  IPropsAPIContext,
  PropsAPIContext
} from "../contexts";
import { useTheme } from "../hooks";
import {
  getNodeConfig,
  getPointByClient,
  getRectHeight,
  getRectWidth,
  IZoomPanSettings
} from "../utils";
import { throttle } from "../utils/throttle";
import { GraphNodeState, ICanvasNode } from "./Graph.interface";

interface IGraphNodeControlPointsProps {
  node: ICanvasNode;
  zoomPanSettings: IZoomPanSettings;
}

const BBOX_PADDING = 15;

let lastClientX = 0;
let lastClientY = 0;
let curWidth = 0;
let curHeight = 0;
let startX = 0;
let startY = 0;

export const GraphNodeControlPoints: React.FunctionComponent<IGraphNodeControlPointsProps> = props => {
  const { node, zoomPanSettings } = props;

  const propsAPIContext = React.useContext<IPropsAPIContext>(PropsAPIContext);

  const graphConfig = React.useContext<IGraphConfig>(GraphConfigContext);
  const nodeConfig = getNodeConfig(node, graphConfig);

  const minWidth = nodeConfig.getMinWidth(node);
  const minHeight = nodeConfig.getMinHeight(node);

  const { theme } = useTheme();

  const handleNWResize = throttle((evt: MouseEvent) => {
    const clientDX = evt.clientX - lastClientX;
    const clientDY = evt.clientY - lastClientY;

    const { x: dx, y: dy } = getPointByClient(
      clientDX,
      clientDY,
      zoomPanSettings.transformMatrix
    );

    const finalDx = Math.min(dx, curWidth - minWidth);
    const finalDy = Math.min(dy, curHeight - minHeight);

    propsAPIContext.propsAPI.updateNode(node.id, {
      x: startX + finalDx,
      y: startY + finalDy,
      width: curWidth - finalDx,
      height: curHeight - finalDy
    });
  }, 50);

  const handleNWResizeEnd = () => {
    window.removeEventListener("mousemove", handleNWResize);
    window.removeEventListener("mouseup", handleNWResizeEnd);
  };

  const handleNResize = throttle((evt: MouseEvent) => {
    const clientDX = evt.clientX - lastClientX;
    const clientDY = evt.clientY - lastClientY;

    const { y: dy } = getPointByClient(
      clientDX,
      clientDY,
      zoomPanSettings.transformMatrix
    );

    const finalDy = Math.min(curHeight - minHeight, dy);

    propsAPIContext.propsAPI.updateNode(node.id, {
      y: startY + finalDy,
      height: curHeight - finalDy
    });
  }, 50);

  const handleNResizeEnd = () => {
    window.removeEventListener("mousemove", handleNResize);
    window.removeEventListener("mouseup", handleNResizeEnd);
  };

  const handleNEResize = throttle((evt: MouseEvent) => {
    const clientDX = evt.clientX - lastClientX;
    const clientDY = evt.clientY - lastClientY;

    const { x: dx, y: dy } = getPointByClient(
      clientDX,
      clientDY,
      zoomPanSettings.transformMatrix
    );

    const finalDx = Math.max(dx, minWidth - curWidth);
    const finalDy = Math.min(dy, curHeight - minHeight);

    propsAPIContext.propsAPI.updateNode(node.id, {
      y: startY + finalDy,
      width: curWidth + finalDx,
      height: curHeight - finalDy
    });
  }, 50);

  const handleNEResizeEnd = () => {
    window.removeEventListener("mousemove", handleNEResize);
    window.removeEventListener("mouseup", handleNEResizeEnd);
  };

  const handleWResize = throttle((evt: MouseEvent) => {
    const clientDX = evt.clientX - lastClientX;
    const clientDY = evt.clientY - lastClientY;

    const { x: dx } = getPointByClient(
      clientDX,
      clientDY,
      zoomPanSettings.transformMatrix
    );

    const finalDx = Math.min(dx, curWidth - minWidth);

    propsAPIContext.propsAPI.updateNode(node.id, {
      x: startX + finalDx,
      width: curWidth - finalDx
    });
  }, 50);

  const handleWResizeEnd = () => {
    window.removeEventListener("mousemove", handleWResize);
    window.removeEventListener("mouseup", handleWResizeEnd);
  };

  const handleEResize = throttle((evt: MouseEvent) => {
    const clientDX = evt.clientX - lastClientX;
    const clientDY = evt.clientY - lastClientY;

    const { x: dx } = getPointByClient(
      clientDX,
      clientDY,
      zoomPanSettings.transformMatrix
    );

    const finalDx = Math.max(dx, minWidth - curWidth);

    propsAPIContext.propsAPI.updateNode(node.id, {
      width: curWidth + finalDx
    });
  }, 50);

  const handleEResizeEnd = () => {
    window.removeEventListener("mousemove", handleEResize);
    window.removeEventListener("mouseup", handleEResizeEnd);
  };

  const handleSWResize = throttle((evt: MouseEvent) => {
    const clientDX = evt.clientX - lastClientX;
    const clientDY = evt.clientY - lastClientY;

    const { x: dx, y: dy } = getPointByClient(
      clientDX,
      clientDY,
      zoomPanSettings.transformMatrix
    );

    const finalDx = Math.min(curWidth - minWidth, dx);
    const finalDy = Math.max(minHeight - curHeight, dy);

    propsAPIContext.propsAPI.updateNode(node.id, {
      x: startX + finalDx,
      width: curWidth - finalDx,
      height: curHeight + finalDy
    });
  }, 50);

  const handleSWResizeEnd = () => {
    window.removeEventListener("mousemove", handleSWResize);
    window.removeEventListener("mouseup", handleSWResizeEnd);
  };

  const handleSResize = throttle((evt: MouseEvent) => {
    const clientDX = evt.clientX - lastClientX;
    const clientDY = evt.clientY - lastClientY;

    const { y: dy } = getPointByClient(
      clientDX,
      clientDY,
      zoomPanSettings.transformMatrix
    );

    const finalDy = Math.max(minHeight - curHeight, dy);

    propsAPIContext.propsAPI.updateNode(node.id, {
      height: curHeight + finalDy
    });
  }, 50);

  const handleSResizeEnd = () => {
    window.removeEventListener("mousemove", handleSResize);
    window.removeEventListener("mouseup", handleSResizeEnd);
  };

  const handleSEResize = throttle((evt: MouseEvent) => {
    const clientDX = evt.clientX - lastClientX;
    const clientDY = evt.clientY - lastClientY;

    const { x: dx, y: dy } = getPointByClient(
      clientDX,
      clientDY,
      zoomPanSettings.transformMatrix
    );

    const finalDx = Math.max(minWidth - curWidth, dx);
    const finalDy = Math.max(minHeight - curHeight, dy);

    propsAPIContext.propsAPI.updateNode(node.id, {
      width: curWidth + finalDx,
      height: curHeight + finalDy
    });
  }, 50);

  const handleSEResizeEnd = () => {
    window.removeEventListener("mousemove", handleSEResize);
    window.removeEventListener("mouseup", handleSEResizeEnd);
  };

  const height = getRectHeight<ICanvasNode>(nodeConfig, node);
  const width = getRectWidth<ICanvasNode>(nodeConfig, node);
  const controlPointsPos = [
    {
      x: node.x - BBOX_PADDING,
      y: node.y - BBOX_PADDING,
      cursor: "nw-resize",
      mouseDownHandler: () => {
        window.addEventListener("mousemove", handleNWResize);
        window.addEventListener("mouseup", handleNWResizeEnd);
      }
    },
    {
      x: node.x + width / 2,
      y: node.y - BBOX_PADDING,
      cursor: "n-resize",
      mouseDownHandler: () => {
        window.addEventListener("mousemove", handleNResize);
        window.addEventListener("mouseup", handleNResizeEnd);
      }
    },
    {
      x: node.x + width + BBOX_PADDING,
      y: node.y - BBOX_PADDING,
      cursor: "ne-resize",
      mouseDownHandler: () => {
        window.addEventListener("mousemove", handleNEResize);
        window.addEventListener("mouseup", handleNEResizeEnd);
      }
    },
    {
      x: node.x - BBOX_PADDING,
      y: node.y + height / 2,
      cursor: "w-resize",
      mouseDownHandler: () => {
        window.addEventListener("mousemove", handleWResize);
        window.addEventListener("mouseup", handleWResizeEnd);
      }
    },
    {
      x: node.x + width + BBOX_PADDING,
      y: node.y + height / 2,
      cursor: "e-resize",
      mouseDownHandler: () => {
        window.addEventListener("mousemove", handleEResize);
        window.addEventListener("mouseup", handleEResizeEnd);
      }
    },
    {
      x: node.x - BBOX_PADDING,
      y: node.y + height + BBOX_PADDING,
      cursor: "sw-resize",
      mouseDownHandler: () => {
        window.addEventListener("mousemove", handleSWResize);
        window.addEventListener("mouseup", handleSWResizeEnd);
      }
    },
    {
      x: node.x + width / 2,
      y: node.y + height + BBOX_PADDING,
      cursor: "s-resize",
      mouseDownHandler: () => {
        window.addEventListener("mousemove", handleSResize);
        window.addEventListener("mouseup", handleSResizeEnd);
      }
    },
    {
      x: node.x + width + BBOX_PADDING,
      y: node.y + height + BBOX_PADDING,
      cursor: "se-resize",
      mouseDownHandler: () => {
        window.addEventListener("mousemove", handleSEResize);
        window.addEventListener("mouseup", handleSEResizeEnd);
      }
    }
  ];

  if (node.state !== GraphNodeState.selected) {
    return null;
  }

  return (
    <g>
      {controlPointsPos.map((controlPoint, index) => {
        const onMouseDown = (evt: React.MouseEvent<SVGCircleElement>): void => {
          evt.preventDefault();
          evt.stopPropagation();

          lastClientX = evt.clientX;
          lastClientY = evt.clientY;

          curHeight = height;
          curWidth = width;

          startX = node.x;
          startY = node.y;

          controlPoint.mouseDownHandler();
        };

        return (
          <circle
            key={index}
            cx={controlPoint.x}
            cy={controlPoint.y}
            r={4}
            stroke={theme.controlPointColor}
            fill="transparent"
            cursor={controlPoint.cursor}
            onMouseDown={onMouseDown}
          />
        );
      })}
    </g>
  );
};
