import * as React from "react";
import { GraphPortEvent } from "../common/GraphEvent.constant";
import { GraphConfigContext, IGraphConfig } from "../contexts";
import { useTheme } from "../hooks";
import { getNodeConfig, getPortPosition, updatePort } from "../utils";
import { IGraphProps } from "./Graph";
import {
  GraphPortState,
  ICanvasData,
  ICanvasNode,
  ICanvasPort,
  TSetData
} from "./Graph.interface";

interface IProps {
  node: ICanvasNode;
  data: ICanvasData;
  graphProps: IGraphProps;
  setData: TSetData;
  onPortEvent(
    node: ICanvasNode,
    port: ICanvasPort,
    eventName: GraphPortEvent
  ): React.EventHandler<React.MouseEvent<SVGGElement>>;
}

export const GraphOneNodePorts: React.FunctionComponent<IProps> = props => {
  const { data, node, graphProps, onPortEvent } = props;
  const setData = props.setData.bind(null);
  const graphConfig = React.useContext<IGraphConfig>(GraphConfigContext);
  const nodeConfig = getNodeConfig(node, graphConfig);
  const { theme } = useTheme();

  if (!nodeConfig.getPorts) {
    return null;
  }

  const ports = nodeConfig.getPorts({
    model: node,
    graphProps,
    setData
  });
  const reversePorts = ports.reverse(); // to make tooltip(hover on port) has a correct order

  const onMouseEnter = (port: ICanvasPort) => (
    evt: React.MouseEvent<SVGGElement>
  ) => {
    if (port.state === GraphPortState.default || !port.state) {
      updatePort(
        node.id,
        port.id,
        { state: GraphPortState.activated },
        setData
      );
    }
    onPortEvent(node, port, GraphPortEvent.onPortMouseEnter)(evt);
  };

  const onMouseLeave = (port: ICanvasPort) => (
    evt: React.MouseEvent<SVGGElement>
  ) => {
    if (port.state === GraphPortState.activated) {
      updatePort(node.id, port.id, { state: GraphPortState.default }, setData);
    }
    onPortEvent(node, port, GraphPortEvent.onPortMouseLeave)(evt);
  };

  return (
    <g>
      {reversePorts.map(p => {
        const portShape = p.shape ? p.shape : "default";
        const portConfig = graphConfig.getPortConfigByName(portShape);

        if (!portConfig || !portConfig.render) {
          return null;
        }

        const pos = getPortPosition(node, p, graphConfig);

        return (
          <g
            key={`${node.id}${p.id}`}
            onClick={onPortEvent(node, p, GraphPortEvent.onPortClick)}
            onDoubleClick={onPortEvent(
              node,
              p,
              GraphPortEvent.onPortDoubleClick
            )}
            onMouseDown={onPortEvent(node, p, GraphPortEvent.onPortMouseDown)}
            onMouseUp={onPortEvent(node, p, GraphPortEvent.onPortMouseUp)}
            onContextMenu={onPortEvent(
              node,
              p,
              GraphPortEvent.onPortContextMenu
            )}
            onMouseEnter={onMouseEnter(p)}
            onMouseLeave={onMouseLeave(p)}
            onMouseMove={onPortEvent(node, p, GraphPortEvent.onPortMouseMove)}
            onMouseOver={onPortEvent(node, p, GraphPortEvent.onPortMouseOver)}
            onMouseOut={onPortEvent(node, p, GraphPortEvent.onPortMouseOut)}
            onFocus={undefined}
            onBlur={undefined}
          >
            {portConfig.render({
              model: p,
              data,
              parentNode: node,
              graphProps,
              setData,
              theme,
              ...pos
            })}
          </g>
        );
      })}
    </g>
  );
};
