import * as React from "react";
import { GraphConfigContext, IGraphConfig } from "../../contexts";
import { usePropsAPI } from "../../hooks";
import {
  adjustedClientPoint,
  getClientByPoint,
  getNodeConfig,
  getRectHeight,
  getRectWidth,
  IZoomPanSettings
} from "../../utils";
import { graphController, GraphState } from "../../utils/graphController";
import { throttle } from "../../utils/throttle";
import { IGraphProps } from "../Graph";
import { ICanvasNode, TSetData } from "../Graph.interface";
import { GraphNode } from "../GraphNode";

interface IProps {
  visible: boolean;
  setData: TSetData;
  graphProps: IGraphProps;
  graphSvgRef: React.RefObject<SVGSVGElement>;
  zoomPanSettings: IZoomPanSettings;
}

let isItemInCanvas = false;

export const GraphItemPanel: React.FunctionComponent<IProps> = props => {
  const propsAPI = usePropsAPI();
  const [draggingNode, setDraggingNode] = React.useState();
  const { graphProps, visible, setData, graphSvgRef, zoomPanSettings } = props;

  const graphConfig = React.useContext<IGraphConfig>(GraphConfigContext);

  const onMouseEnter = () => {
    isItemInCanvas = true;
  };
  const onMouseLeave = () => {
    isItemInCanvas = false;
  };

  React.useEffect(() => {
    if (graphController.getGraphState() === GraphState.addingNode && visible) {
      window.addEventListener("mousemove", handleDragItem);
      window.addEventListener("mouseup", handleDragItemEnd);
    }
  }, [visible]);

  const handleDragItem = throttle((evt: MouseEvent): void => {
    const shape = graphController.getDraggingShape();
    const { model: node } = graphController.getDraggingItem();
    const [adjustedX, adjustedY] = adjustPosition(evt.clientX, evt.clientY);

    setDraggingNode({
      ...node,
      shape: shape || node.shape,
      x: adjustedX,
      y: adjustedY
    });
  }, 20);

  const handleDragItemEnd = (evt: MouseEvent) => {
    graphController.setGraphState(GraphState.default);

    if (isItemInCanvas) {
      const { model: node } = graphController.getDraggingItem();
      const [adjustedX, adjustedY] = adjustPosition(evt.clientX, evt.clientY);

      propsAPI.addNode({
        ...node,
        x: adjustedX,
        y: adjustedY
      });
    }

    window.removeEventListener("mousemove", handleDragItem);
    window.removeEventListener("mouseup", handleDragItemEnd);
  };

  const adjustPosition = (
    clientX: number,
    clientY: number
  ): [number, number] => {
    const { model: node } = graphController.getDraggingItem();
    const nodeConfig = getNodeConfig(node, graphConfig);
    const rectWidth = getRectWidth<ICanvasNode>(nodeConfig, node);
    const rectHeight = getRectHeight<ICanvasNode>(nodeConfig, node);

    // get transformed width and height
    const { x: realX, y: realY } = getClientByPoint(
      node.x,
      node.y,
      zoomPanSettings.transformMatrix
    );
    const { x: realX2, y: realY2 } = getClientByPoint(
      node.x + rectWidth,
      node.y + rectHeight,
      zoomPanSettings.transformMatrix
    );
    const realWidth = realX2 - realX;
    const realHeight = realY2 - realY;

    const [adjustedX, adjustedY] = adjustedClientPoint(
      clientX - realWidth / 2,
      clientY - realHeight / 2,
      graphSvgRef,
      zoomPanSettings
    );
    return [adjustedX, adjustedY];
  };

  const {
    left = 0,
    top = 0,
    width = "100%",
    height = "100%"
  } = graphSvgRef.current ? graphSvgRef.current.style : {};

  const svgStyle: React.CSSProperties = {
    position: "absolute",
    left,
    top,
    width,
    height
  };

  return (
    <>
      {visible && (
        <svg
          onMouseLeave={onMouseLeave}
          onMouseEnter={onMouseEnter}
          style={svgStyle}
          preserveAspectRatio="xMidYMid meet"
        >
          <g transform={`matrix(${zoomPanSettings.transformMatrix.join(" ")})`}>
            {graphController.getGraphState() === GraphState.addingNode &&
              draggingNode && (
                <GraphNode
                  node={draggingNode}
                  setData={setData}
                  graphProps={graphProps}
                />
              )}
          </g>
        </svg>
      )}
    </>
  );
};
