import * as React from "react";
import { usePropsAPI, useTheme } from "../hooks";
import {
  getClientByPoint,
  getTransformedPointByRealPoint,
  getZoomFitMatrix
} from "../utils";
import { ReadonlyGraph } from "./ReadonlyGraph";

interface IMiniMapProps {
  style?: React.CSSProperties;
}

export const Minimap: React.FunctionComponent<IMiniMapProps> = props => {
  const propsAPI = usePropsAPI();
  const data = propsAPI.getData();
  const svgRef = React.createRef<SVGSVGElement>();

  const [indicatorPathD, setIndicatorPathD] = React.useState<string>();

  const minimapContainerStyle: React.CSSProperties = {
    height: "320px",
    width: "320px",
    ...props.style
  };

  const onClick = (evt: React.MouseEvent) => {
    if (!svgRef.current) {
      return;
    }

    const containerEl = svgRef.current.closest<HTMLDivElement>(
      ".react-dag-editor-container"
    );

    if (!containerEl) {
      return;
    }

    const { height, width, top, left } = containerEl.getBoundingClientRect();

    const point = getTransformedPointByRealPoint(
      evt.clientX - left,
      evt.clientY - top,
      getZoomFitMatrix(data, height, width)
    );

    propsAPI.scrollIntoView(point.x, point.y);
  };

  React.useEffect(() => {
    if (!svgRef.current) {
      return () => {
        // do nothing
      };
    }

    const containerEl = svgRef.current.closest<HTMLDivElement>(
      ".react-dag-editor-container"
    );

    if (!containerEl) {
      return () => {
        // do nothing
      };
    }

    const { height, width } = containerEl.getBoundingClientRect();
    const boundaryPoints = propsAPI.getBoundaryPoints();

    const { x: startX, y: startY } = getClientByPoint(
      boundaryPoints.minX,
      boundaryPoints.minY,
      getZoomFitMatrix(data, height, width)
    );

    const { x: endX, y: endY } = getClientByPoint(
      boundaryPoints.maxX,
      boundaryPoints.maxY,
      getZoomFitMatrix(data, height, width)
    );
    const boxWidth = endX - startX;
    const boxHeight = endY - startY;

    setIndicatorPathD(
      `m${startX} ${startY} v ${boxHeight} h ${boxWidth} v${-boxHeight} h ${-boxWidth}`
    );

    return () => {
      // do nothing
    };
  });

  const { theme } = useTheme();

  return (
    <div className="minimap-container" style={minimapContainerStyle}>
      <ReadonlyGraph data={data} svgRef={svgRef} />
      <svg
        onClick={onClick}
        style={{
          position: "absolute",
          top: 0,
          left: 0,
          height: "100%",
          width: "100%"
        }}
      >
        <path d={indicatorPathD} stroke={theme.primaryColor} fill="none" />
      </svg>
    </div>
  );
};
