import * as React from "react";
import {
  defaultAnnotation,
  DefaultClipboard,
  defaultPort,
  DefaultStorage,
  line,
  rect,
  withDefaultPortsPosition
} from "../built-in";
import {
  ContextMenuConfig,
  ContextMenuConfigContext,
  GraphConfig,
  GraphConfigContext,
  IPanelConfig,
  IPanelState,
  ITheme,
  PanelContext,
  PropsAPI,
  PropsAPIContext,
  ThemeContext
} from "../contexts";
import { DefaultErrorBoundary } from "./DefaultErrorBoundary";
import { RegisterAnnotation } from "./RegisterAnnotation";
import { RegisterClipboard } from "./RegisterClipboard";
import { RegisterEdge } from "./RegisterEdge";
import { RegisterNode } from "./RegisterNode";
import { RegisterPort } from "./RegisterPort";

export interface IReactDagEditorProps {
  style?: React.CSSProperties;
  className?: string;
  errorBoundary?: React.ComponentClass | React.FunctionComponent;
  theme?: ITheme;
}

export const ReactDagEditor: React.FunctionComponent<IReactDagEditorProps> = props => {
  const ErrorBoundary = props.errorBoundary
    ? props.errorBoundary
    : DefaultErrorBoundary;

  const clipboardStorage = window.localStorage
    ? window.localStorage
    : new DefaultStorage();

  const [panelState, setPanelState] = React.useState<IPanelState>({});
  const defaultTheme = {
    controlPointColor: "#333333",
    primaryColor: "#0078D4",
    defaultColor: "#CCCCCC",
    borderColor: "#B3B0AD",
    defaultBorderColor: "#FFFFFF",
    unConnectableBgColor: "#E1DFDD",
    defaultBackgroundColor: "#FFFFFF",
    annotationBgColor: "#FFFFE0",
    portStroke: "#ccc",
    portFill: "#fff",
    connectedPortColor: "gray",
    nodeActivateFill: "#ffffff",
    nodeActivateStroke: "#0078D4",
    nodeFill: "#ffffff",
    nodeStroke: "#cccccc",
    contextMenuBackground: "#FFFFFF",
    contextMenuBorder: "#E1DFDD",
    contextMenuHoverBackground: "rgba(0, 120, 212, 0.05)",
    fontColor: "#000000",
    canvasBackground: "#EDEDED",
    edgeColor: "#ccc",
    edgeColorSelected: "#015cda",
    ...props.theme
  };
  const [theme, setTheme] = React.useState<ITheme>(defaultTheme);

  return (
    <ErrorBoundary>
      <GraphConfigContext.Provider value={new GraphConfig()}>
        <ContextMenuConfigContext.Provider value={new ContextMenuConfig()}>
          <PropsAPIContext.Provider value={{ propsAPI: new PropsAPI() }}>
            <ThemeContext.Provider value={{ theme, setTheme }}>
              <PanelContext.Provider
                value={{
                  configList: new Map<string, IPanelConfig>(),
                  state: panelState,
                  setState: setPanelState
                }}
              >
                <RegisterNode
                  name="rect"
                  config={withDefaultPortsPosition(rect)}
                />
                <RegisterEdge name="line" config={line} />
                <RegisterPort name="default" config={defaultPort} />
                <RegisterClipboard
                  clipboard={new DefaultClipboard(clipboardStorage)}
                />
                <RegisterAnnotation
                  name="annotation"
                  config={defaultAnnotation}
                />
                <div style={props.style} className={props.className}>
                  {props.children}
                </div>
              </PanelContext.Provider>
            </ThemeContext.Provider>
          </PropsAPIContext.Provider>
        </ContextMenuConfigContext.Provider>
      </GraphConfigContext.Provider>
    </ErrorBoundary>
  );
};
