import * as React from "react";
import { createUseStyles } from "react-jss";
import { usePropsAPI } from "../hooks";
import { ICanvasNode } from "./Graph.interface";

export interface ISearchBoxProps {
  nodeSearchCriteria(searchKey: string, node: ICanvasNode): boolean;
  renderResult?(node: ICanvasNode): React.ReactNode;
}

export interface ISearchBoxState {
  searchText: string;
  results: ICanvasNode[];
  selectedIndex: number;
}

export const SearchBox: React.FunctionComponent<ISearchBoxProps> = props => {
  const propsApi = usePropsAPI();

  const [state, setState] = React.useState<ISearchBoxState>({
    searchText: "",
    results: [],
    selectedIndex: 0
  });

  const useStyles = createUseStyles({
    searchTextInput: {
      border: "none",
      borderBottom: "1px solid #ccc",
      "&:focus": {
        outline: "none"
      }
    },
    searchResultContainer: {
      zIndex: 9999,
      position: "absolute",
      listStyle: "none",
      paddingLeft: 0,
      marginTop: 0
    },
    searchResultItem: {
      "$:focus": {
        outline: "1px solid blue"
      }
    }
  });
  const classes = useStyles();

  const onSearchTextChange: React.ChangeEventHandler<HTMLInputElement> = evt => {
    const { value } = evt.target;

    const { nodes } = propsApi.getData();
    const results = value
      ? nodes.filter(n => props.nodeSearchCriteria(value, n))
      : [];

    propsApi.markSearchResults(results.map(r => r.id));

    setState(prevState => {
      return {
        ...prevState,
        searchText: value,
        results
      };
    });
  };

  const onKeyDown: React.KeyboardEventHandler = evt => {
    if (state.results.length === 0) {
      return;
    }

    if (evt.key === "ArrowDown") {
      setState(prevState => {
        const selectedIndex = Math.min(
          prevState.results.length - 1,
          prevState.selectedIndex + 1
        );

        propsApi.setSelected(prevState.results[selectedIndex].id);

        return {
          ...prevState,
          selectedIndex
        };
      });
    } else if (evt.key === "ArrowUp") {
      setState(prevState => {
        const selectedIndex = Math.max(0, prevState.selectedIndex - 1);

        propsApi.setSelected(prevState.results[selectedIndex].id);

        return {
          ...prevState,
          selectedIndex
        };
      });
    }
  };

  const renderResult = props.renderResult
    ? props.renderResult.bind(null)
    : (result: ICanvasNode) => (
        <div className="search-result-item">{result.name || ""}</div>
      );

  return (
    <div onKeyDown={onKeyDown} role="application">
      <div className="search-text-input-container">
        <input
          type="text"
          value={state.searchText}
          onChange={onSearchTextChange}
          className={classes.searchTextInput}
        />
      </div>
      <ul
        className={`${classes.searchResultContainer} search-result-container`}
      >
        {state.results.map((result, index) => {
          const border =
            index === state.selectedIndex ? "2px solid blue" : "none";
          const liStyle: React.CSSProperties = {
            border
          };

          const onLIClick: React.MouseEventHandler = evt => {
            evt.stopPropagation();

            setState(prevState => {
              propsApi.setSelected(prevState.results[index].id);

              return {
                ...prevState,
                selectedIndex: index
              };
            });
          };

          return (
            <li key={index} style={liStyle} onClick={onLIClick} role="button">
              {renderResult(result)}
            </li>
          );
        })}
      </ul>
    </div>
  );
};
