import * as React from "react";
import { IGraphClipBoard } from "../built-in/defaultClipboard";
import {
  ICanvasAnnotation,
  ICanvasData,
  ICanvasEdge,
  ICanvasNode,
  ICanvasPort,
  IGraphProps,
  TSetData
} from "../components";
import { ITheme } from "./ThemeContext";

interface IItemConfigArgs<T> {
  model: T;
  graphProps: IGraphProps;
  setData: TSetData;
  theme: ITheme;
}

export interface INodeDrawArgs extends IItemConfigArgs<ICanvasNode> {}

export interface IRectConfig<T> {
  portShape?: string;
  getPorts?(args: Omit<INodeDrawArgs, "theme">): ICanvasPort[];
  render(args: IItemConfigArgs<T>): React.ReactNode;
  getStyle?(rect: T, theme: ITheme): React.CSSProperties;
  getMinWidth(rect: T): number;
  getMinHeight(rect: T): number;
}

export interface IEdgeDrawArgs extends IItemConfigArgs<ICanvasEdge> {
  x1: number;
  y1: number;
  x2: number;
  y2: number;
}

export interface IEdgeConfig {
  render(args: IEdgeDrawArgs): React.ReactNode;
  getStyle?(edge: ICanvasEdge, theme: ITheme): React.CSSProperties;
}

export interface IPortDrawArgs extends IItemConfigArgs<ICanvasPort> {
  data: ICanvasData;
  parentNode: ICanvasNode;
  x: number;
  y: number;
}

export interface IPortConfig {
  render(args: IPortDrawArgs): React.ReactNode;
  getIsConnectable(port: ICanvasPort, data: ICanvasData): boolean | undefined;
  getStyle?(
    port: ICanvasPort,
    data: ICanvasData,
    theme: ITheme
  ): React.CSSProperties;
}

export interface IGraphConfig {
  registerNode(name: string, config: IRectConfig<ICanvasNode>): void;
  getNodeConfigByName(name: string): IRectConfig<ICanvasNode> | void;
  registerEdge(name: string, config: IEdgeConfig): void;
  getEdgeConfigByName(name: string): IEdgeConfig | void;
  registerPort(name: string, config: IPortConfig): void;
  getPortConfigByName(name: string): IPortConfig | void;
  registerAnnotation(
    name: string,
    config: IRectConfig<ICanvasAnnotation>
  ): void;
  getAnnotationConfigByName(
    name: string
  ): IRectConfig<ICanvasAnnotation> | void;
  registerClipboard(clipboard: IGraphClipBoard): void;
  getClipboard(): IGraphClipBoard;
}

interface IRectConfigItem<T> {
  name: string;
  config: IRectConfig<T>;
}

interface IEdgeConfigItem {
  name: string;
  config: IEdgeConfig;
}

interface IPortConfigItem {
  name: string;
  config: IPortConfig;
}

export class GraphConfig implements IGraphConfig {
  private readonly nodeConfigList: Array<IRectConfigItem<ICanvasNode>> = [];
  private readonly edgeConfigList: IEdgeConfigItem[] = [];
  private readonly portConfigList: IPortConfigItem[] = [];
  private readonly annotationConfigList: Array<
    IRectConfigItem<ICanvasAnnotation>
  > = [];
  private clipboard?: IGraphClipBoard;

  public registerNode(name: string, config: IRectConfig<ICanvasNode>): void {
    this.nodeConfigList.push({ name, config });
  }

  public getNodeConfigByName(name: string): IRectConfig<ICanvasNode> | void {
    const item = this.nodeConfigList.find(config => config.name === name);

    return item ? item.config : undefined;
  }

  public registerEdge(name: string, config: IEdgeConfig): void {
    this.edgeConfigList.push({ name, config });
  }

  public getEdgeConfigByName(name: string): IEdgeConfig | void {
    const item = this.edgeConfigList.find(config => config.name === name);

    return item ? item.config : undefined;
  }

  public registerPort(name: string, config: IPortConfig): void {
    this.portConfigList.push({ name, config });
  }

  public getPortConfigByName(name: string): IPortConfig | void {
    const item = this.portConfigList.find(config => config.name === name);

    return item ? item.config : undefined;
  }

  public registerAnnotation(
    name: string,
    config: IRectConfig<ICanvasAnnotation>
  ): void {
    this.annotationConfigList.push({ name, config });
  }
  public getAnnotationConfigByName(
    name: string
  ): IRectConfig<ICanvasAnnotation> | void {
    const item = this.annotationConfigList.find(config => config.name === name);

    return item ? item.config : undefined;
  }

  public registerClipboard(clipboard: IGraphClipBoard): void {
    this.clipboard = clipboard;
  }

  public getClipboard(): IGraphClipBoard {
    if (!this.clipboard) {
      throw new Error("No clipboard registered!");
    }

    return this.clipboard;
  }
}

export const GraphConfigContext = React.createContext<IGraphConfig>(
  new GraphConfig()
);
