import * as React from "react";
import { COPIED_NODE_SPACING } from "../built-in/defaultClipboard";
import {
  GraphNodeState,
  ICanvasAnnotation,
  ICanvasData,
  ICanvasEdge,
  ICanvasNode,
  ICanvasPort,
  TSetData
} from "../components";
import {
  addAnnotation,
  addEdge,
  addNode,
  adjustedClientPoint,
  deleteSelectedItems,
  filterSelectedItems,
  getBoundaryPoints,
  getClientByPoint,
  getEntityById,
  updateNode
} from "../utils";
import { dataStore } from "../utils/undoRedo";
import {
  IZoomPanSettings,
  pan,
  resetZoom,
  scrollIntoView,
  TSetZoomPanSettings,
  zoom,
  zoomTo,
  zoomToFit
} from "../utils/zoomAndPan";
import { IGraphConfig } from "./GraphConfigContext";
import { IPanelContext } from "./PanelContext";

export interface IPropsAPI {
  init(
    _setData: TSetData,
    _getData: () => ICanvasData,
    svgRef: React.RefObject<SVGSVGElement>,
    zoomPanSettings: IZoomPanSettings,
    setZoomPanSettings: TSetZoomPanSettings,
    graphConfig: IGraphConfig,
    panelContext: IPanelContext
  ): void;
  getNodeById(id: string): ICanvasNode | void;
  read(data: ICanvasData): void;
  save(): ICanvasData;
  setData(data: ICanvasData): void;
  getData(): ICanvasData;
  addNode(node: ICanvasNode): void;
  setSelected(nodeId: string): void;
  addEdge(edge: ICanvasEdge): void;
  updateNode(id: string, patch: Partial<ICanvasNode>): ICanvasNode | false;
  deletePort(portId: string, nodeId: string): void;
  addPort<P extends ICanvasPort>(nodeId: string, port: P): void;
  delete(): void;
  zoom(scale: number): void;
  pan(dx: number, dy: number): void;
  resetZoom(): void;
  zoomTo(scale: number): void;
  zoomToFit(): void;
  copy(): void;
  paste(x?: number, y?: number): void;
  addAnnotation(annotation: ICanvasAnnotation): void;
  undo(): void;
  redo(): void;
  markSearchResults(matchesIds: string[]): void;
  // tslint:disable-next-line: no-any
  openSidePanel(name: string, data?: any): void;
  dismissSidePanel(): void;
  getVisiblePanelName(): string | void;
  scrollIntoView(x: number, y: number): void;
  getBoundaryPoints(): {
    minX: number;
    minY: number;
    maxX: number;
    maxY: number;
  };
}

export interface IPropsAPIContext {
  propsAPI: IPropsAPI;
}

export class PropsAPI implements IPropsAPI {
  private _setData: TSetData;
  private _getData: () => ICanvasData;
  private _svgRef: React.RefObject<SVGSVGElement>;
  private _zoomPanSettings: IZoomPanSettings;
  private _setZoomPanSettings: React.Dispatch<
    React.SetStateAction<IZoomPanSettings>
  >;
  private _graphConfig: IGraphConfig;
  private _panelContext: IPanelContext;

  public init(
    _setData: TSetData,
    _getData: () => ICanvasData,
    svgRef: React.RefObject<SVGSVGElement>,
    zoomPanSettings: IZoomPanSettings,
    setZoomPanSettings: TSetZoomPanSettings,
    graphConfig: IGraphConfig,
    panelContext: IPanelContext
  ): void {
    this._setData = _setData;
    this._getData = _getData;
    this._svgRef = svgRef;
    this._zoomPanSettings = zoomPanSettings;
    this._setZoomPanSettings = setZoomPanSettings;
    this._graphConfig = graphConfig;
    this._panelContext = panelContext;
  }

  public getNodeById(id: string): ICanvasNode | void {
    const data = this._getData();

    return getEntityById(id, data.nodes);
  }

  public setData(data: ICanvasData): void {
    this._setData(data);
  }

  public read(data: ICanvasData): void {
    // tslint:disable-next-line: no-console
    console.warn('Deprecating. Please use "setData" instead');

    this.setData(data);
  }

  public getData(): ICanvasData {
    return this._getData();
  }

  public save(): ICanvasData {
    // tslint:disable-next-line: no-console
    console.warn('Deprecating. Please use "getData" instead');

    return this.getData();
  }

  public addNode(node: ICanvasNode): void {
    addNode(node, this._setData);
  }

  public addEdge(edge: ICanvasEdge): void {
    addEdge(edge, this._setData);
  }

  public updateNode(
    id: string,
    patch: Partial<ICanvasNode>
  ): ICanvasNode | false {
    const currentNode = this.getNodeById(id);

    if (!currentNode) {
      return false;
    }

    updateNode(id, patch, this._setData);

    return {
      ...currentNode,
      ...patch
    };
  }

  public deletePort(portId: string, nodeId: string): void {
    this._setData(prevData => {
      return {
        ...prevData,
        nodes: prevData.nodes.map(n => {
          if (n.id === nodeId && n.ports) {
            return { ...n, ports: n.ports.filter(p => p.id !== portId) };
          }

          return n;
        }),
        edges: prevData.edges.filter(
          e =>
            !(
              (e.target === nodeId && e.targetPortId === portId) ||
              (e.source === nodeId && e.sourcePortId === portId)
            )
        )
      };
    });
  }

  public addPort<P extends ICanvasPort>(nodeId: string, port: P): void {
    this._setData(prevData => {
      return {
        ...prevData,
        nodes: prevData.nodes.map(n => {
          if (n.id === nodeId) {
            return {
              ...n,
              ports: [...(n.ports || []), port]
            };
          }

          return n;
        })
      };
    });
  }

  public setSelected(nodeId: string): void {
    const data = this._getData();
    const nextData = {
      ...data,
      nodes: data.nodes.map(n => {
        if (n.id === nodeId) {
          return {
            ...n,
            state: GraphNodeState.selected
          };
        }

        if (n.state === GraphNodeState.selected) {
          return {
            ...n,
            state: GraphNodeState.default
          };
        }

        return n;
      })
    };

    this.setData(nextData);
  }

  public delete(): void {
    deleteSelectedItems(this._setData);
  }

  public zoom(scale: number): void {
    zoom(scale, this._svgRef, this._setZoomPanSettings);
  }

  public pan(dx: number, dy: number): void {
    pan(dx, dy, this._setZoomPanSettings);
  }

  public resetZoom(): void {
    resetZoom(this._setZoomPanSettings);
  }

  public zoomTo(scale: number): void {
    zoomTo(scale, this._svgRef, this._setZoomPanSettings);
  }

  public zoomToFit(): void {
    zoomToFit(this._getData(), this._setZoomPanSettings, this._svgRef);
  }

  public copy(): void {
    const selectedData = filterSelectedItems(this._getData());
    const clipboard = this._graphConfig.getClipboard();

    clipboard.write(selectedData);
  }
  public paste(x?: number, y?: number): void {
    const clipboard = this._graphConfig.getClipboard();

    const pasteData = clipboard.read();

    if (!pasteData) {
      return;
    }

    let pasteNodes = pasteData.nodes;

    if (x && y) {
      const [adjustedX, adjustedY] = adjustedClientPoint(
        x,
        y,
        this._svgRef,
        this._zoomPanSettings
      );

      let dx: number;
      let dy: number;

      pasteNodes = pasteData.nodes.map((n, idx) => {
        // (dx,dy) are same for all copied nodes
        if (idx === 0) {
          dx = adjustedX - n.x;
          dy = adjustedY - n.y;
        }

        return {
          ...n,
          x: dx ? n.x - COPIED_NODE_SPACING + dx : n.x,
          y: dy ? n.y - COPIED_NODE_SPACING + dy : n.y
        };
      });
    }

    const prevData = this._getData();

    const nextData = {
      ...prevData,
      nodes: [...prevData.nodes, ...pasteNodes],
      edges: [...prevData.edges, ...pasteData.edges]
    };

    dataStore.record(nextData);

    this._setData(nextData);
  }

  public addAnnotation(annotation: ICanvasAnnotation): void {
    const [adjustedX, adjustedY] = adjustedClientPoint(
      annotation.x,
      annotation.y,
      this._svgRef,
      this._zoomPanSettings
    );

    const transformedAnnotation = {
      ...annotation,
      x: adjustedX,
      y: adjustedY
    };
    addAnnotation(transformedAnnotation, this._setData);
  }

  public redo(): void {
    const data = dataStore.redo();

    if (data) {
      this.setData(data);
    }
  }

  public undo(): void {
    const data = dataStore.undo();

    if (data) {
      this.setData(data);
    }
  }

  public markSearchResults(matchesIds: string[]): void {
    const data = this.getData();
    const nextData = {
      ...data,
      nodes: data.nodes.map(n => {
        return {
          ...n,
          isInSearchResults: matchesIds.indexOf(n.id) > -1
        };
      })
    };

    this.setData(nextData);
  }

  // tslint:disable-next-line: no-any
  public openSidePanel(name: string, data?: any): void {
    const prevPanelName = this._panelContext.state.name;
    const prevPanelConfig = prevPanelName
      ? this._panelContext.configList.get(prevPanelName)
      : null;
    const nextPanelConfig = this._panelContext.configList.get(name);

    this._panelContext.setState({
      name,
      data
    });

    if (prevPanelConfig && prevPanelConfig.panelDidDismiss) {
      prevPanelConfig.panelDidDismiss();
    }

    if (nextPanelConfig && nextPanelConfig.panelDidOpen) {
      nextPanelConfig.panelDidOpen();
    }
  }

  public dismissSidePanel(): void {
    const prevPanelName = this._panelContext.state.name;
    const prevPanelConfig = prevPanelName
      ? this._panelContext.configList.get(prevPanelName)
      : null;

    this._panelContext.setState({
      name: undefined,
      data: null
    });

    if (prevPanelConfig && prevPanelConfig.panelDidDismiss) {
      prevPanelConfig.panelDidDismiss();
    }
  }

  public getVisiblePanelName(): string | void {
    return this._panelContext.state?.name;
  }

  public scrollIntoView(x: number, y: number): void {
    const { x: clientX, y: clientY } = getClientByPoint(
      x,
      y,
      this._zoomPanSettings.transformMatrix
    );

    scrollIntoView(
      clientX,
      clientY,
      this._svgRef,
      this._setZoomPanSettings,
      true
    );
  }

  public getBoundaryPoints(): {
    minX: number;
    minY: number;
    maxX: number;
    maxY: number;
  } {
    return getBoundaryPoints(this._svgRef, this._zoomPanSettings);
  }
}

export const PropsAPIContext = React.createContext<IPropsAPIContext>({
  propsAPI: new PropsAPI()
});
