import * as React from "react";
import { TSetData } from "../components";
import { IGraphConfig } from "../contexts";
import { deleteSelectedItems, filterSelectedItems } from "../utils";
import { dataStore } from "../utils/undoRedo";
import { useWindowEvent } from "./useWindowEvent";

export interface IUseKeyboardHandlersArgs {
  setIsSpaceHeld: React.Dispatch<React.SetStateAction<boolean>>;
  setData: TSetData;
  getGraphConfig(): IGraphConfig;
  getIsCanvasFocused(): boolean;
}

const keyDownHandlerMap = new Map<string, (evt: KeyboardEvent) => void>();
const keyUpHandlerMap = new Map<string, (evt: KeyboardEvent) => void>();

export const useKeyboardHandlers = (args: IUseKeyboardHandlersArgs) => {
  React.useMemo(() => {
    const spaceKeyDownHandler = (evt: KeyboardEvent) => {
      if (evt.repeat) {
        return;
      }

      args.setIsSpaceHeld.call(null, true);
    };
    keyDownHandlerMap.set(" ", spaceKeyDownHandler);

    const deleteKeyDownHandler = (evt: KeyboardEvent) => {
      evt.preventDefault();
      evt.stopPropagation();
      deleteSelectedItems(args.setData.bind(null));
    };
    keyDownHandlerMap.set("Delete", deleteKeyDownHandler);

    const cKeyDownHandler = (evt: KeyboardEvent) => {
      if (evt.ctrlKey) {
        evt.preventDefault();
        evt.stopPropagation();
        // A trick here: the setData call does not update any data at all. Just to make sure the data is the latest one.
        args.setData(data => {
          const selectedData = filterSelectedItems(data);
          const clipboard = args.getGraphConfig().getClipboard();

          clipboard.write(selectedData);

          return data;
        });
      }
    };
    keyDownHandlerMap.set("c", cKeyDownHandler);

    const vKeyDownHandler = (evt: KeyboardEvent) => {
      if (evt.ctrlKey) {
        evt.preventDefault();
        evt.stopPropagation();

        const clipboard = args.getGraphConfig().getClipboard();

        const pasteData = clipboard.read();

        if (!pasteData) {
          return;
        }

        args.setData(prevData => {
          const nextData = {
            ...prevData,
            nodes: [...prevData.nodes, ...pasteData.nodes],
            edges: [...prevData.edges, ...pasteData.edges]
          };

          dataStore.record(nextData);

          return nextData;
        });
      }
    };
    keyDownHandlerMap.set("v", vKeyDownHandler);

    const zKeyDownHandler = (evt: KeyboardEvent) => {
      if (evt.ctrlKey) {
        evt.preventDefault();
        evt.stopPropagation();

        const nextData = dataStore.undo();

        if (nextData) {
          args.setData(nextData);
        }
      }
    };

    keyDownHandlerMap.set("z", zKeyDownHandler);

    const yKeyDownHandler = (evt: KeyboardEvent) => {
      if (evt.ctrlKey) {
        evt.preventDefault();
        evt.stopPropagation();

        const nextData = dataStore.redo();

        if (nextData) {
          args.setData(nextData);
        }
      }
    };

    keyDownHandlerMap.set("y", yKeyDownHandler);

    const spaceKeyUpHandler = (evt: KeyboardEvent) => {
      evt.preventDefault();
      evt.stopPropagation();

      args.setIsSpaceHeld.call(null, false);
    };
    keyUpHandlerMap.set(" ", spaceKeyUpHandler);
  }, []);

  useWindowEvent("keydown", (evt: KeyboardEvent) => {
    if (!args.getIsCanvasFocused()) {
      return;
    }

    const handler = keyDownHandlerMap.get(evt.key);

    if (handler) {
      handler.call(null, evt);
    }
  });

  useWindowEvent("keyup", (evt: KeyboardEvent) => {
    if (!args.getIsCanvasFocused()) {
      return;
    }

    const handler = keyUpHandlerMap.get(evt.key);

    if (handler) {
      evt.preventDefault();
      evt.stopPropagation();

      handler.call(null, evt);
    }
  });
};
