// tslint:disable: export-name

import { v4 as uuid } from "uuid";
import {
  GraphEdgeState,
  GraphNodeState,
  GraphPortConnectState,
  GraphPortState,
  ICanvasAnnotation,
  ICanvasData,
  ICanvasEdge,
  ICanvasNode,
  ICanvasPort,
  IGraphProps,
  TSetData
} from "../components";
import { IGraphConfig } from "../contexts";
import { getPortPositionByPortId } from "./getPortPosition";
import { dataStore } from "./undoRedo";

export function getEntityById<T extends { id: string }>(
  id: string,
  data?: T[]
): T {
  const matches = (data || []).filter(n => n.id === id);

  if (!matches) {
    throw new Error(`invalid id ${id}`);
  }

  return matches[0];
}

export const updateNode = (
  nodeId: string,
  patch: Partial<ICanvasNode>,
  setData: TSetData
): void => {
  updateEntity<ICanvasNode>(nodeId, "nodes", patch, setData);
};

export const updateAnnotation = (
  annotationId: string,
  patch: Partial<ICanvasAnnotation>,
  setData: TSetData
): void => {
  updateEntity<ICanvasAnnotation>(annotationId, "annotations", patch, setData);
};

export const updateEdge = (
  edgeId: string,
  patch: Partial<ICanvasEdge>,
  setData: TSetData
): void => {
  updateEntity<ICanvasEdge>(edgeId, "edges", patch, setData);
};

export function updateEntity<T extends { id: string }>(
  entityId: string,
  entityType: keyof ICanvasData,
  patch: Partial<T>,
  setData: TSetData
): void {
  setData(data => {
    const key = entityType as string;

    const newData = data[key]?.map((n: T) => {
      if (n.id === entityId) {
        return {
          ...n,
          ...patch
        };
      }
      return { ...n };
    });

    return {
      ...data,
      [key]: newData
    };
  });
}

export const moveNode = (
  nodeId: string,
  dx: number,
  dy: number,
  setData: TSetData
): void => {
  setData(data => {
    const nodes = data.nodes.map(n => {
      if (n.id === nodeId) {
        return {
          ...n,
          x: n.x + dx,
          y: n.y + dy
        };
      }

      return { ...n };
    });

    const nextData = { ...data, nodes };

    dataStore.debouncedRecord(nextData);

    return nextData;
  });
};

export const updatePort = (
  nodeId: string,
  portId: string,
  patch: Partial<ICanvasPort>,
  setData: TSetData
): void => {
  setData(data => {
    const nodes = data.nodes.map(n => {
      if (n.id === nodeId) {
        if (n.ports) {
          const ports = n.ports.map(p => {
            if (p.id === portId) {
              return {
                ...p,
                ...patch
              };
            }
            return { ...p };
          });
          return { ...n, ports };
        }
        return { ...n };
      }
      return { ...n };
    });

    return {
      ...data,
      nodes
    };
  });
};

export const getNearestConnectablePort = (
  node: ICanvasNode,
  clientX: number,
  clientY: number,
  graphConfig: IGraphConfig,
  graphProps: IGraphProps,
  setData: TSetData,
  transformMatrix: [number, number, number, number, number, number]
): ICanvasPort | undefined => {
  let minDistance = Infinity;
  let nearestPort;

  // use "setData" is to get the latest data
  setData(data => {
    if (!node.ports) {
      return { ...data };
    }
    node.ports.forEach(port => {
      if (isConnectable(graphConfig, port, data)) {
        const portPos = getPortPositionByPortId(
          node,
          port.id,
          graphConfig,
          graphProps,
          setData,
          transformMatrix
        );
        const diffX = clientX - portPos.x;
        const diffY = clientY - portPos.y;
        const distance = Math.pow(diffX * diffX + diffY * diffY, 0.5);
        if (distance < minDistance) {
          minDistance = distance;
          nearestPort = { ...port, position: [portPos.x, portPos.y] };
        }
      }
    });

    return { ...data };
  });

  return nearestPort;
};

export const isConnectable = (
  graphConfig: IGraphConfig,
  port?: ICanvasPort,
  data?: ICanvasData
) => {
  if (!port || !data) {
    return false;
  }
  const portShape = port.shape ? port.shape : "default";
  const portConfig = graphConfig.getPortConfigByName(portShape);
  return portConfig ? portConfig.getIsConnectable(port, data) : false;
};

export const addEdgeByPosition = (
  nodes: ICanvasNode[],
  source: string,
  sourcePortId: string,
  clientX: number,
  clientY: number,
  graphConfig: IGraphConfig,
  graphProps: IGraphProps,
  setData: TSetData,
  transformMatrix: [number, number, number, number, number, number]
): void => {
  const targetInfo = getPortByPosition(
    nodes,
    clientX,
    clientY,
    graphConfig,
    graphProps,
    setData,
    transformMatrix
  );

  if (
    !targetInfo ||
    (source === targetInfo.node.id && sourcePortId === targetInfo.port.id)
  ) {
    setData(resetPortsState);

    return;
  }
  const { node: targetNode, port: targetPort } = targetInfo;

  addEdge(
    {
      source,
      sourcePortId,
      target: targetNode?.id || "",
      targetPortId: targetPort?.id || "",
      id: uuid()
    },
    setData
  );
};

export const addEdge = (edge: ICanvasEdge, setData: TSetData): void => {
  setData(data => {
    const edges = [...data.edges];
    const { source, sourcePortId, target, targetPortId } = edge;
    if (
      edge &&
      source &&
      sourcePortId &&
      target &&
      targetPortId &&
      !isEdgeExist(edges, edge)
    ) {
      edges.push(edge);
    }

    const nextData = resetPortsState({
      ...data,
      edges
    });

    dataStore.record(nextData);

    return nextData;
  });
};

const getPortByPosition = (
  nodes: ICanvasNode[],
  clientX: number,
  clientY: number,
  graphConfig: IGraphConfig,
  graphProps: IGraphProps,
  setData: TSetData,
  transformMatrix: [number, number, number, number, number, number]
):
  | {
      node: ICanvasNode;
      port: ICanvasPort;
    }
  | undefined => {
  for (const node of nodes) {
    const { ports = [] } = node;

    for (const port of ports) {
      const portPos = getPortPositionByPortId(
        node,
        port.id,
        graphConfig,
        graphProps,
        setData,
        transformMatrix
      );

      if (isMouseHoverOnThePort(clientX, clientY, portPos)) {
        return {
          node,
          port
        };
      }
    }
  }
  return undefined;
};

const isMouseHoverOnThePort = (
  clientX: number,
  clientY: number,
  portPos: { x: number; y: number }
): boolean => {
  const minX = portPos.x - 10;
  const minY = portPos.y - 10;
  const maxX = portPos.x + 20;
  const maxY = portPos.y + 20;
  if (clientX > minX && clientX < maxX && clientY > minY && clientY < maxY) {
    return true;
  }
  return false;
};

const isEdgeExist = (edges: ICanvasEdge[], edge: ICanvasEdge): boolean => {
  for (const e of edges) {
    if (
      e.source === edge.source &&
      e.target === edge.target &&
      e.sourcePortId === edge.sourcePortId &&
      e.targetPortId === edge.targetPortId
    ) {
      return true;
    }
  }
  return false;
};

export const resetPortsState = (data: ICanvasData): ICanvasData => {
  const { nodes } = data;

  const nextNodes = nodes.map(curNode => {
    if (!curNode.ports) {
      return curNode;
    }

    const nextPorts = curNode.ports.map(curPort => {
      return {
        ...curPort,
        state: GraphPortState.default
      };
    });

    return {
      ...curNode,
      ports: nextPorts
    };
  });

  return {
    ...data,
    nodes: nextNodes
  };
};

export const normalizePortsConnectState = (data: ICanvasData): ICanvasData => {
  const { edges, nodes } = data;

  const nextNodes = nodes.map(curNode => {
    if (!curNode.ports) {
      return curNode;
    }

    const nextPorts = curNode.ports.map(curPort => {
      const isConnectedAsSource = edges.some(
        e => e.sourcePortId === curPort.id && e.source === curNode.id
      );

      const isConnectedAsTarget = edges.some(
        e => e.targetPortId === curPort.id && e.target === curNode.id
      );

      if (isConnectedAsSource) {
        return {
          ...curPort,
          connectState: GraphPortConnectState.connectedAsSource
        };
      }

      if (isConnectedAsTarget) {
        return {
          ...curPort,
          connectState: GraphPortConnectState.connectedAsTarget
        };
      }

      return {
        ...curPort,
        connectState: undefined
      };
    });

    return {
      ...curNode,
      ports: nextPorts
    };
  });

  return {
    ...data,
    nodes: nextNodes
  };
};

export const filterSelectedItems = (data: ICanvasData): ICanvasData => {
  const selectedNodeIds = getSelectedEntityIds<ICanvasNode>(data.nodes);

  if (selectedNodeIds.length === 0) {
    return { nodes: [], edges: [] };
  }

  return {
    nodes: data.nodes.filter(n => selectedNodeIds.indexOf(n.id) >= 0),
    edges: data.edges.filter(
      e =>
        e.state === GraphEdgeState.selected ||
        (selectedNodeIds.indexOf(e.source) >= 0 &&
          selectedNodeIds.indexOf(e.target) >= 0)
    )
  };
};

export const deleteSelectedItems = (setData: TSetData): void => {
  setData(data => {
    const { nodes, edges, annotations } = data;
    const selectedNodeIds = getSelectedEntityIds<ICanvasNode>(nodes);
    const selectedEdgeIds = getSelectedEntityIds<ICanvasEdge>(edges);
    const selectedAnnotationsIds = getSelectedEntityIds<ICanvasAnnotation>(
      annotations
    );

    if (
      !selectedNodeIds.length &&
      !selectedEdgeIds.length &&
      !selectedAnnotationsIds.length
    ) {
      return data;
    }

    const nextData = resetPortsState({
      ...data,
      nodes: nodes.filter(n => selectedNodeIds.indexOf(n.id) < 0),
      edges: edges.filter(
        e =>
          e.state !== GraphEdgeState.selected &&
          selectedNodeIds.indexOf(e.source) < 0 &&
          selectedNodeIds.indexOf(e.target) < 0
      ),
      annotations: annotations?.filter(
        a => selectedAnnotationsIds.indexOf(a.id) < 0
      )
    });

    dataStore.record(nextData);

    return nextData;
  });
};

export const addNode = (node: ICanvasNode, setData: TSetData): void => {
  addEntity(node, setData, "nodes");
};

export const addAnnotation = (
  annotation: ICanvasAnnotation,
  setData: TSetData
): void => {
  addEntity(annotation, setData, "annotations");
};

const addEntity = (
  entity: ICanvasNode | ICanvasAnnotation,
  setData: TSetData,
  entityType: keyof Pick<ICanvasData, "nodes" | "annotations">
): void => {
  setData(data => {
    const nextData = {
      ...data,
      [entityType]: [...(data[entityType] || []), entity]
    };

    dataStore.record(nextData);

    return nextData;
  });
};

export function getSelectedEntityIds<
  T extends { id: string; state?: GraphNodeState | GraphEdgeState }
>(entities?: T[]): string[] {
  return (entities || [])
    .filter(n => n.state === GraphNodeState.selected)
    .map(n => n.id);
}
