import { ICanvasData, ICanvasNode } from "../components";
import { getClientByPoint } from "./transformMatrix";

export interface IZoomPanSettings {
  transformMatrix: [number, number, number, number, number, number];
}

export type TSetZoomPanSettings = React.Dispatch<
  React.SetStateAction<IZoomPanSettings>
>;

export const zoomTo = (
  scale: number,
  svgRef: React.RefObject<SVGSVGElement>,
  setZoomPanSettings: TSetZoomPanSettings
): void => {
  if (!svgRef.current) {
    return;
  }

  const containerEl = svgRef.current.closest<HTMLDivElement>(
    ".react-dag-editor-container"
  );

  if (!containerEl) {
    return;
  }
  const { width, height } = containerEl.getBoundingClientRect();
  const dx = (width * (1 - scale)) / 2;
  const dy = (height * (1 - scale)) / 2;

  setZoomPanSettings(prevState => {
    return {
      ...prevState,
      transformMatrix: [scale, 0, 0, scale, dx, dy]
    };
  });
};

export const zoom = (
  scale: number,
  svgRef: React.RefObject<SVGSVGElement>,
  setZoomPanSettings: TSetZoomPanSettings
): void => {
  if (!svgRef.current) {
    return;
  }

  const containerEl = svgRef.current.closest<HTMLDivElement>(
    ".react-dag-editor-container"
  );

  if (!containerEl) {
    return;
  }
  const { width, height } = containerEl.getBoundingClientRect();
  const dx = (width * (1 - scale)) / 2;
  const dy = (height * (1 - scale)) / 2;

  setZoomPanSettings(prevState => {
    return {
      ...prevState,
      transformMatrix: [
        prevState.transformMatrix[0] * scale,
        prevState.transformMatrix[1] * scale,
        prevState.transformMatrix[2] * scale,
        prevState.transformMatrix[3] * scale,
        prevState.transformMatrix[4] * scale + dx,
        prevState.transformMatrix[5] * scale + dy
      ]
    };
  });
};

export const pan = (
  dx: number,
  dy: number,
  setZoomPanSettings: TSetZoomPanSettings
): void => {
  setZoomPanSettings(prevState => {
    return {
      ...prevState,
      transformMatrix: [
        prevState.transformMatrix[0],
        prevState.transformMatrix[1],
        prevState.transformMatrix[2],
        prevState.transformMatrix[3],
        prevState.transformMatrix[4] + dx,
        prevState.transformMatrix[5] + dy
      ]
    };
  });
};

export const resetZoom = (setZoomPanSettings: TSetZoomPanSettings): void => {
  setZoomPanSettings(prevState => {
    return {
      ...prevState,
      transformMatrix: [1, 0, 0, 1, 0, 0]
    };
  });
};

export const getZoomFitMatrix = (
  data: ICanvasData,
  height: number,
  width: number
): IZoomPanSettings["transformMatrix"] => {
  const { nodes } = data;
  const minNodeX = Math.min(...nodes.map(n => n.x));
  const maxNodeX = Math.max(...nodes.map(n => n.x));
  const minNodeY = Math.min(...nodes.map(n => n.y));
  const maxNodeY = Math.max(...nodes.map(n => n.y));
  const scale = Math.min(
    (width * 0.9) / (maxNodeX - minNodeX),
    (height * 0.9) / (maxNodeY - minNodeY)
  );

  const dx = -scale * minNodeX;
  const dy = -scale * minNodeY;

  return [scale, 0, 0, scale, dx, dy];
};

export const zoomToFit = (
  data: ICanvasData,
  setZoomPanSettings: TSetZoomPanSettings,
  svgRef: React.RefObject<SVGSVGElement>
): void => {
  if (!svgRef.current) {
    return;
  }

  const containerEl = svgRef.current.closest<HTMLDivElement>(
    ".react-dag-editor-container"
  );

  if (!containerEl) {
    return;
  }

  const { width, height } = containerEl.getBoundingClientRect();

  setZoomPanSettings(prevState => {
    return {
      ...prevState,
      transformMatrix: getZoomFitMatrix(data, height, width)
    };
  });
};

export const scrollIntoView = (
  x: number,
  y: number,
  svgRef: React.RefObject<SVGSVGElement>,
  setZoomPanSettings: TSetZoomPanSettings,
  alwaysScroll?: boolean
): void => {
  if (!svgRef.current) {
    return;
  }

  const containerEl = svgRef.current.closest<HTMLDivElement>(
    ".react-dag-editor-container"
  );

  if (!containerEl) {
    return;
  }
  const { width, height, top, left } = containerEl.getBoundingClientRect();
  const shouldScroll =
    x < left || x > width + left || y < top || y > height + top;

  if (!shouldScroll && !alwaysScroll) {
    return;
  }

  setZoomPanSettings(prevState => {
    const dx = width / 2 + left - x;

    const dy = height / 2 + top - y;

    return {
      ...prevState,
      transformMatrix: [
        prevState.transformMatrix[0],
        prevState.transformMatrix[1],
        prevState.transformMatrix[2],
        prevState.transformMatrix[3],
        prevState.transformMatrix[4] + dx,
        prevState.transformMatrix[5] + dy
      ]
    };
  });
};

export const focusNode = (
  node: ICanvasNode,
  svgRef: React.RefObject<SVGSVGElement>,
  zoomAndPanSettings: IZoomPanSettings,
  setZoomPanSettings: TSetZoomPanSettings
): void => {
  const { x, y } = getClientByPoint(
    node.x,
    node.y,
    zoomAndPanSettings.transformMatrix
  );
  scrollIntoView(x, y, svgRef, setZoomPanSettings);
};
