import { storiesOf } from "@storybook/react";
import { initializeIcons } from "@uifabric/icons";
import * as React from "react";

import { Spinner } from "office-ui-fabric-react";
import { Graph, ICanvasEdge, ICanvasNode, ReactDagEditor } from "../src";
import { CustomErrorBoundary } from "./components/CustomErrorBoundary";
import { DemoGraph } from "./components/DemoGraph";
import { getSample1Data } from "./components/getSample1Data";
import { getSample4Data } from "./components/getSample4Data";

initializeIcons();

const generateData = (nodesCount: number) => {
  const nodes: ICanvasNode[] = [];
  const edges: ICanvasEdge[] = [];

  for (let i = 0; i < nodesCount; i++) {
    nodes.push({
      name: `node-${i}`,
      x: i * 130 + 25,
      y: i * 130 + 25,
      id: `node${i}`,
      shape: "module",
      data: {
        comment: "",
        iconName: "ModelingView",
        statusIcon: <Spinner />
      },
      ports: [
        {
          id: "0",
          name: "port-0",
          isInputDisabled: false,
          isOutputDisabled: true,
          shape: "modulePort"
        },
        {
          id: "1",
          name: "port-1",
          isInputDisabled: true,
          isOutputDisabled: false,
          shape: "modulePort"
        }
      ]
    });

    if (i > 0) {
      edges.push({
        source: `node${i - 1}`,
        sourcePortId: "0",
        target: `node${i}`,
        targetPortId: "1",
        id: String(i)
      });
    }
  }

  return { nodes, edges };
};

storiesOf("ReactDagEditor", module)
  .add("Sample1", () => {
    return <DemoGraph initData={getSample1Data()} />;
  })
  .add("Sample4", () => {
    return <DemoGraph initData={getSample4Data()} />;
  })
  .add("Sample1-uncontrolled", () => {
    return (
      <ReactDagEditor>
        <Graph
          data={getSample1Data()}
          style={{ height: "100vh", width: "100%", backgroundColor: "#EDEDED" }}
        />
      </ReactDagEditor>
    );
  })
  .add("500 nodes", () => {
    return <DemoGraph initData={generateData(500)} />;
  })
  .add("10000 nodes", () => {
    return <DemoGraph initData={generateData(10000)} />;
  })
  .add("Node not draggable", () => {
    return (
      <ReactDagEditor>
        <Graph data={generateData(10)} isNodeDraggingDisabled={true} />
      </ReactDagEditor>
    );
  })
  .add("Default Error Boundary: invalid source id in edge", () => {
    const data = generateData(10);

    data.edges.push({
      id: "11",
      source: "99",
      sourcePortId: "0",
      target: "1",
      targetPortId: "0"
    });

    return (
      <ReactDagEditor>
        <Graph data={data} />
      </ReactDagEditor>
    );
  })
  .add("Default Error Boundary: invalid target id in edge", () => {
    const data = generateData(10);

    data.edges.push({
      id: "11",
      source: "1",
      sourcePortId: "0",
      target: "99",
      targetPortId: "0"
    });

    return (
      <ReactDagEditor>
        <Graph data={data} />
      </ReactDagEditor>
    );
  })
  .add("Custom Error Boundary: invalid source id in edge", () => {
    const data = generateData(10);

    data.edges.push({
      id: "11",
      source: "99",
      sourcePortId: "0",
      target: "1",
      targetPortId: "0"
    });

    return (
      <ReactDagEditor errorBoundary={CustomErrorBoundary}>
        <Graph data={data} />
      </ReactDagEditor>
    );
  })
  .add("Custom Error Boundary: invalid target id in edge", () => {
    const data = generateData(10);

    data.edges.push({
      id: "11",
      source: "1",
      sourcePortId: "0",
      target: "99",
      targetPortId: "0"
    });

    return (
      <ReactDagEditor errorBoundary={CustomErrorBoundary}>
        <Graph data={data} />
      </ReactDagEditor>
    );
  });
