import * as React from "react";
import {
  IAccessibleTreeStyles,
  ITreeNode,
  ReactAccessibleTree
} from "react-accessible-tree";
import { v4 as uuid } from "uuid";
import { Item } from "../../src";
import * as moduleData from "../data/modules.json";
import { getIconByCategory } from "./getSample1Data";
import { ICanvasPortCustomized } from "./modulePort";

interface IItem {
  name?: string;
  id?: string;
  ports: ICanvasPortCustomized[];
  category: string;
  // nodeType: canvasNodeType;
  // isGlobalDataset?: boolean;
  // globalDatasetId?: string;
  // datasetInfo?: IPipelineDatasetNode;
  // moduleInfo?: IPipelineModuleNode;
  // description?: string;
  // parameters?: IStructuredInterfaceParameter | object;
  // isImportDataModule?: boolean;
  // isExportDataModule?: boolean;
}

enum tagLevel {
  Primary = 1,
  Secondary = 2,
  NotTag = 0
}

export const ItemPanel: React.FunctionComponent = () => {
  const [treeData, setTreeData] = React.useState<ITreeNode[]>([]);
  const treeRoot: ITreeNode = {
    title: "dummyRoot",
    children: [],
    searchKeys: [],
    id: "dummyRoot",
    expanded: false
  };

  const transformToSortTreeData = (
    data: IItem[],
    expandAll?: boolean
  ): ITreeNode[] => {
    treeRoot.children = [];

    data.forEach(m => {
      const levels = m.category.split("/");

      const item = (
        <Item key={m.id} model={m}>
          {m.name}
        </Item>
      );
      const searchKeys = [m.name ? m.name : ""];
      ensureTreePath(
        levels,
        {
          title: item,
          searchKeys,
          id: m.id || m.name || ""
        },
        treeRoot
      );
    });

    return expandAll ? expandAllChildren(treeRoot.children) : treeRoot.children;
  };

  const ensureTreePath = (
    levels: string[],
    leafNode: Omit<ITreeNode, "children">,
    curNode: ITreeNode
  ): void => {
    if (levels.length === 0) {
      curNode.children.push({
        title: leafNode.title,
        children: [],
        searchKeys: leafNode.searchKeys,
        id: leafNode.id,
        ariaLabel: leafNode.searchKeys[0],
        extra: leafNode.extra
      });

      return;
    }

    const curLevel = levels[0];
    const curChildren = curNode.children.filter(c => c.id === curLevel);

    if (curChildren.length === 0) {
      const curTagLevel = getTagLevel(curLevel);
      const treeTitle =
        curTagLevel === tagLevel.NotTag ? (
          <div className="title-node ms-fontSize-12">
            <span>{curLevel}</span>
          </div>
        ) : curTagLevel === tagLevel.Primary ? (
          <div className="tag-node tag-node-primary title-node ms-fontSize-14">
            <span>{curLevel}</span>
          </div>
        ) : (
          <div className="tag-node title-node ms-fontSize-12">
            <span>{curLevel}</span>
          </div>
        );
      curTagLevel === tagLevel.NotTag
        ? curNode.children.push({
            title: treeTitle,
            children: [],
            searchKeys: [curLevel],
            id: curLevel,
            ariaLabel: curLevel
          })
        : curNode.children.push({
            title: treeTitle,
            children: [],
            searchKeys: [curLevel],
            id: curLevel,
            isTag: true,
            ariaLabel: curLevel
          });
    }
    const curChild = curNode.children.filter(c => c.id === curLevel)[0];
    ensureTreePath(levels.slice(1), leafNode, curChild);
  };

  const expandAllChildren = (data: ITreeNode[]): ITreeNode[] => {
    return data.map(item => ({
      ...item,
      expanded: true,
      children: expandAllChildren(item.children)
    }));
  };

  const getTagLevel = (title: string): number => {
    switch (title) {
      case "Regression":
      case "Clustering":
      case "Classification":
      case "My Datasets":
      case "Samples":
        return 2;
      case "Manage modules":
        return 1;
      default:
        return 0;
    }
  };

  // data: IModuleEntity[]
  const formatTreeData = (data: any[]): IItem[] => {
    const moduleList: IItem[] = data.map(m => {
      const ports: ICanvasPortCustomized[] = m.structuredInterface
        ? [
            ...(m.structuredInterface.inputs
              ? m.structuredInterface.inputs.map((input: any) => ({
                  id: uuid(),
                  shape: "modulePort",
                  name: input.name || "",
                  // inputDataFormats: input.dataTypeIdsList || [],
                  // outputDataFormats: [],
                  isInputDisabled: false,
                  isOutputDisabled: true,
                  label: input.label
                  // isOptional: input.isOptional
                }))
              : []),
            ...(m.structuredInterface.outputs
              ? m.structuredInterface.outputs.map((output: any) => ({
                  id: uuid(),
                  shape: "modulePort",
                  name: output.name || "",
                  // inputDataFormats: [],
                  // outputDataFormats: output.dataTypeId ? [output.dataTypeId] : [],
                  isInputDisabled: true,
                  isOutputDisabled: false,
                  label: output.label
                }))
              : [])
          ]
        : [];

      return {
        id: m.id,
        name: m.name,
        shape: "module",
        ports,
        category: m.category,
        data: {
          iconName: getIconByCategory(m.category)
        }
      };
    });

    return [...moduleList];
  };

  const modulesToDisplay = formatTreeData(moduleData);

  if (treeData.length === 0) {
    setTreeData(transformToSortTreeData(modulesToDisplay, true));
  }

  const onChange = (nextData: ITreeNode[]) => {
    setTreeData(nextData);
  };
  const onItemKeyDown = () => {
    //
  };

  const treeViewStyles: IAccessibleTreeStyles = {
    root: {
      paddingLeft: 0,
      paddingRight: 0,
      paddingTop: 0,
      marginTop: 0,
      overflowY: "auto",
      willChange: "transform",
      height: "100vh",
      minWidth: 300,
      maxWidth: 300
    },
    // group: { paddingLeft: 0 },
    item: { listStyle: "none" }
  };
  return (
    <ReactAccessibleTree
      treeData={treeData}
      onChange={onChange}
      styles={treeViewStyles}
      onItemKeyDown={onItemKeyDown}
    />
  );
};
