import {
  CommandBar,
  ICommandBarItemProps,
  Icon,
  IStackStyles,
  Slider,
  Stack
} from "office-ui-fabric-react";
import * as React from "react";
import * as uuid from "uuid";
import {
  getRectHeight,
  getRectWidth,
  ICanvasNode,
  SearchBox,
  usePropsAPI
} from "../../src";
import { moduleNode } from "./moduleNode";

const formatSliderValue = (value: number) =>
  `${parseInt(value.toString(), 10)}%`;

const nodeSearchCriteria = (searchKey: string, node: ICanvasNode): boolean => {
  if (!node.name) {
    return false;
  }

  return node.name.toLowerCase().indexOf(searchKey.toLowerCase()) > -1;
};

const getNodeStyle = (node: ICanvasNode): IStackStyles => {
  return {
    root: {
      height: getRectHeight<ICanvasNode>(moduleNode, node),
      padding: 8,
      backgroundColor: "#fff",
      borderRadius: 4,
      cursor: "move",
      border: `2px solid #ffffff`
    }
  };
};

const renderStatusIcon = (node: ICanvasNode): React.ReactNode => {
  if (!node.data || !node.data.statusIcon) {
    return null;
  }

  return node.data.statusIcon;
};

const renderSearchResult = (result: ICanvasNode): React.ReactNode => {
  const containerStyles = getNodeStyle(result);
  const mainContainerStyles = {
    root: {
      height: 58
    }
  };
  const iconStyles = {
    root: {
      marginRight: 8,
      color: "#0078D4"
    }
  };
  const comment = result.data && result.data.comment;
  const iconName = result.data && result.data.iconName;

  return (
    <div
      style={{
        height: getRectHeight<ICanvasNode>(moduleNode, result),
        width: getRectWidth<ICanvasNode>(moduleNode, result),
        background: "#ccc",
        padding: "5px"
      }}
    >
      <Stack styles={containerStyles}>
        <Stack
          horizontal={true}
          verticalAlign="center"
          styles={mainContainerStyles}
        >
          <Stack.Item>
            <Icon
              iconName={iconName}
              styles={iconStyles}
              style={{ userSelect: "none" }}
            />
          </Stack.Item>
          <Stack.Item
            grow={true}
            disableShrink={true}
            styles={{ root: { width: 200, marginBottom: 5 } }}
          >
            <div style={{ userSelect: "none" }}>{result.name}</div>
          </Stack.Item>
          <Stack.Item styles={{ root: { userSelect: "none" } }}>
            {renderStatusIcon(result)}
          </Stack.Item>
        </Stack>
        <div style={{ color: "#ccc", fontSize: 12 }}>{comment}</div>
      </Stack>
    </div>
  );
};

interface IProps {
  scale: number;
  isMinimapVisible: boolean;
  setIsMinimapVisible(isMinimapVisible: boolean): void;
}

export const Toolbar: React.FunctionComponent<IProps> = props => {
  const propsAPI = usePropsAPI();

  const onSliderChange = (value: number) => {
    propsAPI.zoomTo(value / 100);
  };

  const items: ICommandBarItemProps[] = [
    {
      key: "addNode",
      text: "addNode",
      ariaLabel: "addNode",
      iconOnly: true,
      iconProps: { iconName: "Add" },
      onClick: () => {
        propsAPI.addNode({
          id: uuid.v4(),
          x: 300,
          y: 300,
          shape: "module",
          name: "new added node"
        });
      }
    },
    {
      key: "addAnnotation",
      text: "addAnnotation",
      ariaLabel: "addAnnotation",
      iconOnly: true,
      iconProps: { iconName: "Comment" },
      onClick: () => {
        propsAPI.addAnnotation({
          id: uuid.v4(),
          x: 400,
          y: 200,
          shape: "annotation"
        });
      }
    },
    {
      key: "delete",
      text: "delete",
      ariaLabel: "delete",
      iconOnly: true,
      iconProps: { iconName: "Delete" },
      onClick: () => {
        propsAPI.delete();
      }
    },
    {
      key: "copy",
      text: "copy",
      ariaLabel: "copy",
      iconOnly: true,
      iconProps: { iconName: "Copy" },
      onClick: () => {
        propsAPI.copy();
      }
    },
    {
      key: "paste",
      text: "paste",
      ariaLabel: "paste",
      iconOnly: true,
      iconProps: { iconName: "Paste" },
      onClick: () => {
        propsAPI.paste();
      }
    },
    {
      key: "Undo",
      text: "Undo",
      ariaLabel: "Undo",
      iconOnly: true,
      iconProps: { iconName: "Undo" },
      onClick: () => {
        propsAPI.undo();
      }
    },
    {
      key: "Redo",
      text: "Redo",
      ariaLabel: "Redo",
      iconOnly: true,
      iconProps: { iconName: "Redo" },
      onClick: () => {
        propsAPI.redo();
      }
    },
    {
      key: "ZoomIn",
      text: "ZoomIn",
      ariaLabel: "ZoomIn",
      iconOnly: true,
      iconProps: { iconName: "ZoomIn" },
      onClick: () => {
        propsAPI.zoom(1.25);
      }
    },
    {
      key: "ZoomOut",
      text: "ZoomOut",
      ariaLabel: "ZoomOut",
      iconOnly: true,
      iconProps: { iconName: "ZoomOut" },
      onClick: () => {
        propsAPI.zoom(0.8);
      }
    },
    {
      key: "ZoomToFit",
      text: "ZoomToFit",
      ariaLabel: "ZoomToFit",
      iconOnly: true,
      iconProps: { iconName: "ZoomToFit" },
      onClick: () => {
        propsAPI.zoomToFit();
      }
    },
    {
      key: "ResetZoom",
      text: "Reset Zoom",
      ariaLabel: "Reset Zoom",
      onClick: () => {
        propsAPI.resetZoom();
      }
    },
    {
      key: "ZoomTo",
      text: "ZoomTo",
      ariaLabel: "ZoomTo",
      onRender: () => {
        const value = props.scale * 100;
        return (
          <Slider
            styles={{ root: { width: 300 } }}
            onChange={onSliderChange}
            min={0}
            max={500}
            step={5}
            value={value}
            valueFormat={formatSliderValue}
          />
        );
      }
    },
    {
      key: "Search",
      text: "Search",
      ariaLabel: "Search",
      onRender: () => {
        return (
          <SearchBox
            nodeSearchCriteria={nodeSearchCriteria}
            renderResult={renderSearchResult}
          />
        );
      }
    },
    {
      key: "Overview",
      text: "Overview",
      ariaLabel: "Overview",
      iconOnly: true,
      iconProps: { iconName: "View" },
      onClick: () => {
        propsAPI.openSidePanel("overview");
      }
    },
    {
      key: "Minimap",
      text: "Minimap",
      ariaLabel: "Minimap",
      iconOnly: true,
      iconProps: { iconName: "World" },
      onClick: () => {
        props.setIsMinimapVisible(!props.isMinimapVisible);
      }
    }
  ];

  return <CommandBar items={items} />;
};
