import { Icon, IStackStyles, Stack } from "office-ui-fabric-react";
import * as React from "react";
import {
  GraphNodeState,
  ICanvasNode,
  IRectConfig,
  ITheme,
  TSetData
} from "../../src";
import { getRectHeight, getRectWidth, updateNode } from "../../src/utils";

const renderStatusIcon = (node: ICanvasNode): React.ReactNode => {
  if (!node.data || !node.data.statusIcon) {
    return null;
  }

  return node.data.statusIcon;
};

const getChangeHandler = (node: ICanvasNode, setData: TSetData) => (
  e: React.ChangeEvent<HTMLTextAreaElement>
) => {
  updateNode(
    node.id,
    { data: { ...node.data, comment: e.target.value } },
    setData
  );
};

export const moduleNode: IRectConfig<ICanvasNode> = {
  getMinHeight(curNode: ICanvasNode): number {
    return (curNode.data && curNode.data.comment) ||
      curNode.state === GraphNodeState.editing
      ? 86
      : 56;
  },
  getMinWidth(): number {
    return 280;
  },
  render(args): React.ReactNode {
    const node = args.model;
    const setData = args.setData.bind(null);
    const isEditing = node.state === GraphNodeState.editing;

    const containerStyles = getNodeStyle(node, args.theme);

    const mainContainerStyles = {
      root: {
        marginTop: 6
      }
    };

    const iconStyles = {
      root: {
        marginRight: 8,
        color: args.theme.primaryColor
      }
    };

    const comment = node.data && node.data.comment;
    const iconName = node.data && node.data.iconName;

    const rectHeight = getRectHeight<ICanvasNode>(moduleNode, node);
    const rectWidth = getRectWidth<ICanvasNode>(moduleNode, node);
    const textAreaStyles = {
      maxWidth: rectWidth - 24,
      maxHeight: rectHeight - 24
    };

    return (
      <foreignObject
        x={node.x}
        y={node.y}
        height={rectHeight}
        width={rectWidth}
        // style={{
        //   boxShadow:
        //     "0px 0.6px 1.8px rgba(0, 0, 0, 0.108), 0px 3.2px 7.2px rgba(0, 0, 0, 0.132)"
        // }}
      >
        <Stack styles={containerStyles}>
          <Stack
            horizontal={true}
            verticalAlign="center"
            styles={mainContainerStyles}
          >
            <Stack.Item>
              <Icon
                iconName={iconName}
                styles={iconStyles}
                style={{ userSelect: "none" }}
              />
            </Stack.Item>
            <Stack.Item
              grow={true}
              disableShrink={true}
              styles={{ root: { width: 200, marginBottom: 5 } }}
            >
              <div
                style={{
                  userSelect: "none",
                  background: node.isInSearchResults ? "yellow" : "#fff"
                }}
              >
                {node.name}
              </div>
            </Stack.Item>
            <Stack.Item styles={{ root: { userSelect: "none" } }}>
              {renderStatusIcon(node)}
            </Stack.Item>
          </Stack>
          {isEditing && (
            <textarea
              value={comment}
              placeholder="Input your comment here"
              onChange={getChangeHandler(node, setData)}
              style={textAreaStyles}
              autoFocus={true}
            />
          )}
          {!isEditing && node.data && node.data.comment && (
            <div style={{ color: "#ccc", fontSize: 12, overflow: "hidden" }}>
              {comment}
            </div>
          )}
        </Stack>
      </foreignObject>
    );
  }
};

const getNodeStyle = (node: ICanvasNode, theme: ITheme): IStackStyles => {
  let borderColor = theme.defaultBorderColor;

  if (node.isInSearchResults) {
    borderColor = "yellow";
  }

  switch (node.state) {
    case GraphNodeState.activated:
      borderColor = theme.nodeActivateStroke;
      break;
    case GraphNodeState.selected:
      borderColor = theme.nodeActivateStroke;
      break;
    default:
  }

  return {
    root: {
      height: getRectHeight<ICanvasNode>(moduleNode, node),
      padding: 8,
      backgroundColor: theme.nodeFill,
      borderRadius: 4,
      cursor: "move",
      border: `2px solid ${borderColor}`
    }
  };
};
